import importlib
import subprocess
import sys
import unittest
from unittest import mock

from kagglehub import package_import
from kagglehub.exceptions import UserCancelledError

from .utils import create_test_cache, parameterized, unauthenticated

UNVERSIONED_HANDLE = "dster/package-test"
VERSIONED_HANDLE = "dster/package-test/versions/1"

YES_INPUTS = ["yes", "y", "YES", "Y", "YeS"]


class TestPackageImport(unittest.TestCase):

    def tearDown(self) -> None:
        names = [name for name in sys.modules if name.startswith("kagglehub_package")]
        for name in names:
            del sys.modules[name]

    @parameterized(*YES_INPUTS)
    def test_package_versioned_succeeds(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                package = package_import(VERSIONED_HANDLE)

                mock_input.assert_called_once()
                self.assertIn("foo", dir(package))
                self.assertEqual("bar", package.foo())

    def test_package_versioned_bypass_confirmation_succeeds(self) -> None:
        with create_test_cache():
            package = package_import(VERSIONED_HANDLE, bypass_confirmation=True)

            self.assertIn("foo", dir(package))
            self.assertEqual("bar", package.foo())

    @parameterized(*YES_INPUTS)
    def test_package_unversioned_succeeds(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                package = package_import(UNVERSIONED_HANDLE)

                mock_input.assert_called_once()
                self.assertIn("Model", dir(package))
                model = package.Model()
                self.assertEqual("this is a test package", model.foo())

    def test_download_private_package_succeeds(self) -> None:
        # Don't require confirmation since we're running as integrationtester.
        with create_test_cache():
            package = package_import("integrationtester/kagglehub-test-private-package")

            self.assertIn("foo", dir(package))
            self.assertEqual("bar", package.foo())

    @parameterized(*YES_INPUTS)
    def test_public_package_with_unauthenticated_succeeds(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                with unauthenticated():
                    package = package_import(UNVERSIONED_HANDLE)

                    mock_input.assert_called_once()
                    self.assertIn("Model", dir(package))
                    model = package.Model()
                    self.assertEqual("this is a test package", model.foo())

    @parameterized("no", "NO", "n", "", "anything but yes")
    def test_package_versioned_aborts(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                with self.assertRaises(UserCancelledError):
                    package_import(VERSIONED_HANDLE)
                mock_input.assert_called_once()

    def test_package_with_dependencies_succeeds(self) -> None:
        with create_test_cache():
            # Make sure pyjokes is not already installed, so we test that the package installs it
            self.assertIsNone(importlib.util.find_spec("pyjokes"))

            package = package_import("dster/joke-teller", bypass_confirmation=True)

            self.assertTrue(len(package.get_joke()) > 0)

            # Uninstall pyjokes now
            del sys.modules["pyjokes"]
            importlib.invalidate_caches()
            subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "pyjokes", "-y"])  # noqa: S603
            self.assertIsNone(importlib.util.find_spec("pyjokes"))



import os

# Ensures tests do not try to read your own kaggle credentials.
os.environ["KAGGLE_CONFIG_DIR"] = "/some-missing-directory"

# All APIs call in tests should go to a local test server.
os.environ["KAGGLE_API_ENVIRONMENT"] = "TEST"



import os
from unittest import mock

import requests

import kagglehub
from kagglehub.config import DISABLE_KAGGLE_CACHE_ENV_VAR_NAME
from kagglehub.env import KAGGLE_DATA_PROXY_URL_ENV_VAR_NAME
from tests.fixtures import BaseTestCase
from tests.utils import clear_imported_kaggle_packages

from .server_stubs import jwt_stub as stub
from .server_stubs import serv

INVALID_ARCHIVE_PACKAGE_HANDLE = "invalid/invalid/invalid/invalid/invalid"
VERSIONED_PACKAGE_HANDLE = "alexisbcook/test-package/versions/1"
UNVERSIONED_PACKAGE_HANDLE = "alexisbcook/test-package"


# Test cases for package_import and get_packet_asset_path.
class TestKaggleCachePackageImport(BaseTestCase):

    def tearDown(self) -> None:
        clear_imported_kaggle_packages()

    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app, KAGGLE_DATA_PROXY_URL_ENV_VAR_NAME, "http://localhost:7778")

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def test_unversioned_package_import(self) -> None:
        with stub.create_env():
            package = kagglehub.package_import(UNVERSIONED_PACKAGE_HANDLE)
            self.assertEqual("kaggle", package.foo())
            self.assertEqual("abcd", package.bar())

    def test_versioned_package_import(self) -> None:
        with stub.create_env():
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)
            self.assertEqual("kaggle", package.foo())
            self.assertFalse(hasattr(package, "bar"))

    def test_kaggle_resolver_skipped(self) -> None:
        with mock.patch.dict(os.environ, {DISABLE_KAGGLE_CACHE_ENV_VAR_NAME: "true"}):
            with stub.create_env():
                # Assert that a ConnectionError is set (uses HTTP server which is not set)
                with self.assertRaises(requests.exceptions.ConnectionError):
                    kagglehub.package_import(UNVERSIONED_PACKAGE_HANDLE)

    def test_versioned_package_import_bad_handle_raises(self) -> None:
        with self.assertRaises(ValueError):
            kagglehub.package_import("bad handle")

    def test_versioned_package_import_returns_same(self) -> None:
        with stub.create_env():
            # Importing the same package a second time returns the same exact package.
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)
            package2 = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)
            self.assertEqual(package, package2)

    def test_versioned_package_import_force_download_returns_different(self) -> None:
        with stub.create_env():
            # Re-importing with force_download re-installs anew.
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)
            package_forced = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE, force_download=True)
            self.assertNotEqual(package, package_forced)

    def test_versioned_package_import_with_force_download_explicitly_false(self) -> None:
        with stub.create_env():
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE, force_download=False)
            self.assertEqual("kaggle", package.foo())
            self.assertFalse(hasattr(package, "bar"))



import hashlib
import inspect
import json
import logging
import os
import sys
import zipfile
from collections.abc import Callable
from datetime import datetime, timezone
from urllib.parse import urlparse

import requests
import requests.auth
from kagglesdk.kaggle_client import KaggleClient
from kagglesdk.kaggle_env import KaggleEnv, get_env, is_in_kaggle_notebook
from kagglesdk.kaggle_http_client import KaggleHttpClient
from packaging.version import parse
from requests.auth import HTTPBasicAuth
from tqdm import tqdm

import kagglehub
from kagglehub.cache import delete_from_cache, get_cached_archive_path
from kagglehub.config import get_kaggle_credentials
from kagglehub.datasets_enums import KaggleDatasetAdapter
from kagglehub.env import (
    KAGGLE_DATA_PROXY_URL_ENV_VAR_NAME,
    is_in_colab_notebook,
    read_kaggle_build_date,
    search_lib_in_call_stack,
)
from kagglehub.exceptions import (
    BackendError,
    ColabEnvironmentError,
    CredentialError,
    DataCorruptionError,
    KaggleEnvironmentError,
    NotFoundError,
    colab_raise_for_status,
)
from kagglehub.handle import CompetitionHandle, ResourceHandle
from kagglehub.integrity import get_md5_checksum_from_response, to_b64_digest, update_hash_from_file

CHUNK_SIZE = 1048576
# The `connect` timeout is the number of seconds `requests` will wait for your client to establish a connection.
# The `read` timeout is the number of seconds that the client will wait BETWEEN bytes sent from the server.
# See: https://requests.readthedocs.io/en/stable/user/advanced/#timeouts
DEFAULT_CONNECT_TIMEOUT = 5  # seconds
DEFAULT_READ_TIMEOUT = 15  # seconds
ACCEPT_RANGE_HTTP_HEADER = "Accept-Ranges"
HTTP_STATUS_404 = 404

already_printed_version_warning = False

_CHECKSUM_MISMATCH_MSG_TEMPLATE = """\
The X-Goog-Hash header indicated a MD5 checksum of:

  {}

but the actual MD5 checksum of the downloaded contents was:

  {}
"""

ADAPTER_TO_USER_AGENT_MAP = {
    KaggleDatasetAdapter.HUGGING_FACE: "hugging_face_data_loader",
    KaggleDatasetAdapter.PANDAS: "pandas_data_loader",
    KaggleDatasetAdapter.POLARS: "polars_data_loader",
}


def get_user_agent() -> str:
    """Identifies the user agent based on available system information.

    Returns:
        str: user agent information.
    """
    user_agents = [f"kagglehub/{kagglehub.__version__}"]
    for lib in ("keras_hub", "keras_nlp", "keras_cv", "keras", "torchtune"):
        lib_info = search_lib_in_call_stack(lib)
        if lib_info is not None:
            user_agents.append(lib_info)
            break

    # Add an appropriate data loader user agent for kagglehub.dataset_load calls
    for frame_info in inspect.stack():
        if frame_info.function != "dataset_load":
            continue
        module = inspect.getmodule(frame_info.frame)

        if not module or not hasattr(module, "__name__") or not module.__name__.startswith("kagglehub.datasets"):
            continue

        # We've confirmed that this is a call to kagglehub.dataset_load. Now figure out which loader was used.
        adapter = frame_info.frame.f_locals["adapter"] if "adapter" in frame_info.frame.f_locals else None
        if adapter and adapter in ADAPTER_TO_USER_AGENT_MAP:
            user_agents.append(ADAPTER_TO_USER_AGENT_MAP[adapter])
            break

    if is_in_kaggle_notebook():
        build_date = read_kaggle_build_date()
        user_agents.append(f"kkb/{build_date}")
    elif is_in_colab_notebook():
        colab_tag = os.getenv("COLAB_RELEASE_TAG")
        runtime_suffix = "-managed" if os.getenv("TBE_RUNTIME_ADDR") else "-unmanaged"
        user_agents.append(f"colab/{colab_tag}{runtime_suffix}")

    return " ".join(user_agents)


def get_response_processor() -> Callable[..., None]:
    return _check_response_version


def _check_response_version(response: requests.Response) -> None:
    global already_printed_version_warning  # noqa: PLW0603
    if already_printed_version_warning:
        return
    latest_version_str = response.headers.get("X-Kaggle-HubVersion")
    if latest_version_str:
        current_version = parse(kagglehub.__version__)
        latest_version = parse(latest_version_str)
        if latest_version > current_version:
            sys.stdout.write(
                "Warning: Looks like you're using an outdated `kagglehub` "
                f"version (installed: {current_version}), please consider "
                f"upgrading to the latest version ({latest_version_str})"
            )
            already_printed_version_warning = True


logger = logging.getLogger(__name__)


def build_kaggle_client() -> KaggleClient:
    credentials = get_kaggle_credentials()
    env = get_env()
    verbose = True if env == KaggleEnv.TEST else False
    if not credentials:
        # Unauthenticated client
        return KaggleClient(
            env=env,
            verbose=verbose,
            user_agent=get_user_agent(),
        )

    return KaggleClient(
        env=env,
        verbose=verbose,
        username=credentials.username,
        password=credentials.key,
        api_token=credentials.api_key,
        user_agent=get_user_agent(),
        response_processor=get_response_processor(),
    )


def download_file(
    response: requests.Response,
    out_file: str,
    resource_handle: ResourceHandle,
    cached_path: str | None = None,
    *,
    extract_auto_compressed_file: bool = False,
) -> bool:
    """
    Issues a call to kaggle api and downloads files. For competition downloads,
    call may return early if local cache is newer than the last time the file was modified.

    Returns:
    bool:  If downloading remote was necessary
    """
    total_size = int(response.headers["Content-Length"]) if "Content-Length" in response.headers else None
    size_read = 0

    if isinstance(resource_handle, CompetitionHandle) and not _download_needed(response, resource_handle, cached_path):
        return False

    expected_md5_hash = get_md5_checksum_from_response(response)
    hash_object = hashlib.md5() if expected_md5_hash else None

    if _is_resumable(response) and total_size and os.path.isfile(out_file):
        size_read = os.path.getsize(out_file)
        update_hash_from_file(hash_object, out_file)

        if size_read == total_size:
            logger.info(f"Download already complete ({size_read} bytes).")
            return True

        logger.info(f"Resuming download from {size_read} bytes ({total_size - size_read} bytes left)...")

        # Send the request again with the 'Range' header.
        with requests.get(
            response.url,  # GCS URL after redirection
            stream=True,
            timeout=(DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT),
            headers={"Range": f"bytes={size_read}-"},
        ) as resumed_response:
            logger.info(f"Resuming download to {out_file} ({size_read}/{total_size}) bytes left.")
            _download_file(resumed_response, out_file, size_read, total_size, hash_object)
    else:
        logger.info(f"Downloading to {out_file}...")
        _download_file(response, out_file, size_read, total_size, hash_object)

    if hash_object:
        actual_md5_hash = to_b64_digest(hash_object)
        if actual_md5_hash != expected_md5_hash:
            os.remove(out_file)  # Delete the corrupted file.
            raise DataCorruptionError(_CHECKSUM_MISMATCH_MSG_TEMPLATE.format(expected_md5_hash, actual_md5_hash))

    # For individual file downloads, the downloaded file may be a zip of the file rather
    # than the file name/type that was requested (e.g. my-big-table.csv.zip and not my-big-table.csv).
    # If that's the case, we should auto-extract it so users get what they expect.
    expected_downloaded_file_name = urlparse(out_file).path.split("/")[-1]
    actual_downloaded_file_name = urlparse(response.url).path.split("/")[-1]
    if (
        extract_auto_compressed_file
        and f"{expected_downloaded_file_name}.zip" == actual_downloaded_file_name
        and zipfile.is_zipfile(out_file)
    ):
        logger.info(f"Extracting zip of {expected_downloaded_file_name}...")
        # Rename the file to match what it really is and make space to write to the expected location
        renamed_auto_compressed_path = f"{out_file}.zip"
        os.rename(out_file, renamed_auto_compressed_path)
        with zipfile.ZipFile(renamed_auto_compressed_path, "r") as f:
            f.extract(expected_downloaded_file_name, os.path.dirname(out_file))
        # We don't need the zipped version anymore
        os.remove(renamed_auto_compressed_path)
    return True


def _is_resumable(response: requests.Response) -> bool:
    return ACCEPT_RANGE_HTTP_HEADER in response.headers and response.headers[ACCEPT_RANGE_HTTP_HEADER] == "bytes"


def _download_file(
    response: requests.Response,
    out_file: str,
    size_read: int,
    total_size: int | None,
    hash_object,  # noqa: ANN001 - no public type for hashlib hash
) -> None:
    open_mode = "ab" if size_read > 0 else "wb"
    if total_size is not None:
        with tqdm(total=total_size, initial=size_read, unit="B", unit_scale=True, unit_divisor=1024) as progress_bar:
            with open(out_file, open_mode) as f:
                for chunk in response.iter_content(CHUNK_SIZE):
                    f.write(chunk)
                    if hash_object:
                        hash_object.update(chunk)
                    size_read = min(total_size, size_read + CHUNK_SIZE)
                    progress_bar.update(len(chunk))
    else:
        with open(out_file, open_mode) as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                f.write(chunk)
                if hash_object:
                    hash_object.update(chunk)


def _download_needed(response: requests.Response, h: ResourceHandle, cached_path: str | None = None) -> bool:
    """
    Determine if a download is needed based on timestamp and cached path.

    Returns:
        bool: download needed.
    """
    if not cached_path:
        return True

    last_modified = response.headers.get("Last-Modified")
    if last_modified is None:
        delete_from_cache(h, cached_path)

        archive_path = get_cached_archive_path(h)
        os.makedirs(os.path.dirname(archive_path), exist_ok=True)

        return True
    else:
        remote_date = datetime.strptime(response.headers["Last-Modified"], "%a, %d %b %Y %H:%M:%S %Z").replace(
            tzinfo=timezone.utc
        )

    file_exists = os.path.exists(cached_path)
    if file_exists:
        local_date = datetime.fromtimestamp(os.path.getmtime(cached_path), tz=timezone.utc)

        download_needed = remote_date >= local_date
        if download_needed:
            delete_from_cache(h, cached_path)

            archive_path = get_cached_archive_path(h)
            os.makedirs(os.path.dirname(archive_path), exist_ok=True)

            return True

        return False

    delete_from_cache(h, cached_path)

    archive_path = get_cached_archive_path(h)
    os.makedirs(os.path.dirname(archive_path), exist_ok=True)

    return True


# These environment variables are set by the Kaggle notebook environment.
KAGGLE_JWT_TOKEN_ENV_VAR_NAME = "KAGGLE_USER_SECRETS_TOKEN"
KAGGLE_DATA_PROXY_TOKEN_ENV_VAR_NAME = "KAGGLE_DATA_PROXY_TOKEN"


class KaggleJwtClient:
    BASE_PATH = "/kaggle-jwt-handler/"

    def __init__(self) -> None:
        self.endpoint = os.getenv(KAGGLE_DATA_PROXY_URL_ENV_VAR_NAME)
        if self.endpoint is None:
            msg = f"The {KAGGLE_DATA_PROXY_URL_ENV_VAR_NAME} should be set."
            raise KaggleEnvironmentError(msg)
        jwt_token = os.getenv(KAGGLE_JWT_TOKEN_ENV_VAR_NAME)
        if jwt_token is None:
            msg = (
                "A JWT Token is required to call Kaggle, "
                f"but none found in environment variable {KAGGLE_JWT_TOKEN_ENV_VAR_NAME}"
            )
            raise CredentialError(msg)

        data_proxy_token = os.getenv(KAGGLE_DATA_PROXY_TOKEN_ENV_VAR_NAME)
        if data_proxy_token is None:
            msg = (
                "A Data Proxy Token is required to call Kaggle, "
                f"but none found in environment variable {KAGGLE_DATA_PROXY_TOKEN_ENV_VAR_NAME}"
            )
            raise CredentialError(msg)

        self.headers = {
            "Content-type": "application/json",
            "X-Kaggle-Authorization": f"Bearer {jwt_token}",
            "X-KAGGLE-PROXY-DATA": data_proxy_token,
        }

    def post(
        self,
        request_name: str,
        data: dict,
        timeout: tuple[float, float] = (DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT),
    ) -> dict:
        url = f"{self.endpoint}{KaggleJwtClient.BASE_PATH}{request_name}"
        with requests.post(
            url,
            headers=self.headers,
            data=bytes(json.dumps(data), "utf-8"),
            timeout=timeout,
        ) as response:
            response.raise_for_status()
            json_response = response.json()
            if "wasSuccessful" not in json_response:
                msg = "'wasSuccessful' field missing from response"
                raise BackendError(msg)
            if not json_response["wasSuccessful"]:
                msg = f"POST failed with: {response.text!s}"
                raise BackendError(msg)
            if "result" not in json_response:
                msg = "'result' field missing from response"
                raise BackendError(msg)
            return json_response["result"]


class ColabClient:
    IS_SUPPORTED_PATH = "/kagglehub/models/is_supported"
    IS_MODEL_SUPPORTED_PATH = "/kagglehub/models/is_supported"
    IS_DATASET_SUPPORTED_PATH = "/kagglehub/datasets/is_supported"
    MOUNT_PATH = "/kagglehub/models/mount"
    MODEL_MOUNT_PATH = "/kagglehub/models/mount"
    DATASET_MOUNT_PATH = "/kagglehub/datasets/mount"
    # TBE_RUNTIME_ADDR serves requests made from `is_supported` and  `_resolve`
    # of ModelColabCacheResolver.
    TBE_RUNTIME_ADDR_ENV_VAR_NAME = "TBE_RUNTIME_ADDR"

    def __init__(self) -> None:
        self.endpoint = os.getenv(self.TBE_RUNTIME_ADDR_ENV_VAR_NAME)
        if self.endpoint is None:
            msg = f"The {self.TBE_RUNTIME_ADDR_ENV_VAR_NAME} should be set."
            raise ColabEnvironmentError(msg)

        self.credentials = get_kaggle_credentials()
        self.headers = {"Content-type": "application/json"}

    def post(self, data: dict, handle_path: str, resource_handle: ResourceHandle | None = None) -> dict | None:
        url = f"http://{self.endpoint}{handle_path}"
        with requests.post(
            url,
            data=json.dumps(data),
            auth=self._get_auth(),
            headers=self.headers,
            timeout=(DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT),
        ) as response:
            if response.status_code == HTTP_STATUS_404:
                raise NotFoundError()
            colab_raise_for_status(response, resource_handle)
            if response.text:
                return response.json()
        return None

    def _get_auth(self) -> requests.auth.AuthBase | None:
        if self.credentials:
            if self.credentials.username and self.credentials.key:
                return HTTPBasicAuth(self.credentials.username, self.credentials.key)
            if self.credentials.api_key:
                return KaggleHttpClient.BearerAuth(self.credentials.api_key)
        return None



import kagglehub
from kagglehub import KaggleDatasetAdapter, PolarsFrameType

lazy_dataset = kagglehub.dataset_load(
    KaggleDatasetAdapter.POLARS,
    "wyattowalsh/basketball",
    "nba.sqlite",
    sql_query="SELECT person_id, player_name FROM draft_history",
)
materialized_dataset = lazy_dataset.collect()
print("Records in NBA dataset: ", materialized_dataset.shape)
print(materialized_dataset.head())

dataset = kagglehub.dataset_load(
    KaggleDatasetAdapter.POLARS,
    "robikscube/textocr-text-extraction-from-images-dataset",
    "annot.parquet",
    polars_frame_type=PolarsFrameType.DATA_FRAME,
    polars_kwargs={"columns": ["image_id", "bbox", "points", "area"]},
)

print("Records in OCR dataset: ", dataset.shape)
print(dataset.head())
print(str(dataset.head()))



import os
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock, patch

from kagglesdk.datasets.types.dataset_api_service import ApiDownloadDatasetRequest

import kagglehub
from kagglehub.clients import build_kaggle_client, download_file, get_user_agent
from kagglehub.exceptions import DataCorruptionError
from kagglehub.handle import DatasetHandle
from tests.fixtures import BaseTestCase

from .server_stubs import kaggle_api_stub as stub
from .server_stubs import serv

DUMMY_HANDLE = DatasetHandle("dummy", "dataset")


class TestKaggleClient(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def test_download_with_integrity_check(self) -> None:
        with TemporaryDirectory() as d:
            out_file = os.path.join(d, "out")

            with build_kaggle_client() as api_client:
                r = ApiDownloadDatasetRequest()
                r.dataset_slug = "no-integrity"

                response = api_client.datasets.dataset_api_client.download_dataset(r)
                download_file(response, out_file, DUMMY_HANDLE)

            with open(out_file) as f:
                self.assertEqual("foo", f.read())

    def test_resumable_download_with_integrity_check(self) -> None:
        with TemporaryDirectory() as d:
            out_file = os.path.join(d, "out")

            # If the out_file already has data, we use the 'Range' header to resume download.
            with open(out_file, "w") as f:
                f.write("fo")  # Should download the remaining "o".

            with self.assertLogs("kagglehub", level="INFO") as cm:
                with build_kaggle_client() as api_client:
                    r = ApiDownloadDatasetRequest()
                    r.dataset_slug = "good"

                    response = api_client.datasets.dataset_api_client.download_dataset(r)
                    download_file(response, out_file, DUMMY_HANDLE)

                    self.assertIn("INFO:kagglehub.clients:Resuming download from 2 bytes (1 bytes left)...", cm.output)

            with open(out_file) as f:
                self.assertEqual("foo", f.read())

    def test_download_no_integrity_check(self) -> None:
        with TemporaryDirectory() as d:
            out_file = os.path.join(d, "out")

            with build_kaggle_client() as api_client:
                r = ApiDownloadDatasetRequest()
                r.dataset_slug = "no-integrity"

                response = api_client.datasets.dataset_api_client.download_dataset(r)
                download_file(response, out_file, DUMMY_HANDLE)

            with open(out_file) as f:
                self.assertEqual("foo", f.read())

    def test_download_corrupted_file_fail_integrity_check(self) -> None:
        with TemporaryDirectory() as d:
            out_file = os.path.join(d, "out")

            with build_kaggle_client() as api_client:
                r = ApiDownloadDatasetRequest()
                r.dataset_slug = "corrupted"

                with self.assertRaises(DataCorruptionError):
                    response = api_client.datasets.dataset_api_client.download_dataset(r)
                    download_file(response, out_file, DUMMY_HANDLE)

            # Assert the corrupted file has been deleted.
            self.assertFalse(os.path.exists(out_file))

    @patch.dict("os.environ", {})
    def test_get_user_agent(self) -> None:
        self.assertEqual(get_user_agent(), f"kagglehub/{kagglehub.__version__}")

    @patch.dict(
        "os.environ", {"KAGGLE_KERNEL_RUN_TYPE": "Interactive", "KAGGLE_DATA_PROXY_URL": "https://dp.kaggle.net"}
    )
    def test_get_user_agent_kkb(self) -> None:
        self.assertEqual(get_user_agent(), f"kagglehub/{kagglehub.__version__} kkb/unknown")

    @patch.dict(
        "os.environ",
        {
            "COLAB_RELEASE_TAG": "release-colab-20230531-060125-RC00",
        },
    )
    @patch("kagglehub.env._is_google_colab", True)
    def test_get_user_agent_colab(self) -> None:
        self.assertEqual(
            get_user_agent(),
            f"kagglehub/{kagglehub.__version__} colab/release-colab-20230531-060125-RC00-unmanaged",
        )

    @patch("importlib.metadata.version")
    @patch("inspect.ismodule")
    @patch("inspect.stack")
    def test_get_user_agent_keras_nlp(
        self, mock_stack: MagicMock, mock_is_module: MagicMock, mock_version: MagicMock
    ) -> None:
        # Mock the call stack and version information.
        mock_stack.return_value = [
            MagicMock(frame=MagicMock(__name__="kagglehub.clients")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models_helpers")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models")),
            MagicMock(frame=MagicMock(__name__="keras_nlp.src.utils.preset_utils")),
            MagicMock(frame=MagicMock(None)),
        ]
        mock_is_module.return_value = True
        mock_version.return_value = "0.15.0"
        self.assertEqual(get_user_agent(), f"kagglehub/{kagglehub.__version__} keras_nlp/0.15.0")

    @patch("importlib.metadata.version")
    @patch("inspect.ismodule")
    @patch("inspect.stack")
    def test_get_user_agent_keras_hub(
        self, mock_stack: MagicMock, mock_is_module: MagicMock, mock_version: MagicMock
    ) -> None:
        # Mock the call stack and version information.
        mock_stack.return_value = [
            MagicMock(frame=MagicMock(__name__="kagglehub.clients")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models_helpers")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models")),
            MagicMock(frame=MagicMock(__name__="keras_hub.src.utils.preset_utils")),
            MagicMock(frame=MagicMock(None)),
        ]
        mock_is_module.return_value = True
        mock_version.return_value = "0.17.0"
        self.assertEqual(get_user_agent(), f"kagglehub/{kagglehub.__version__} keras_hub/0.17.0")

    @patch("importlib.metadata.version")
    @patch("inspect.ismodule")
    @patch("inspect.stack")
    def test_get_user_agent_torch_tune(
        self, mock_stack: MagicMock, mock_is_module: MagicMock, mock_version: MagicMock
    ) -> None:
        # Mock the call stack and version information.
        mock_stack.return_value = [
            MagicMock(frame=MagicMock(__name__="kagglehub.clients")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models_helpers")),
            MagicMock(frame=MagicMock(__name__="kagglehub.models")),
            MagicMock(frame=MagicMock(__name__="torchtune.src.utils.preset_utils")),
            MagicMock(frame=MagicMock(None)),
        ]
        mock_is_module.return_value = True
        mock_version.return_value = "0.18.0"
        self.assertEqual(get_user_agent(), f"kagglehub/{kagglehub.__version__} torchtune/0.18.0")



import os
import sqlite3
from collections.abc import Callable
from typing import Any

# WARNING: This module is intended to be imported only at runtime, with
# specific error handling to inform users that they need to install the
# appropriate `polars-datasets` extras in order to use these methods. Adding
# any new dependencies here must correspond with an addition to the
# list of `polars-datasets` optional dependencies in pyproject.toml.
import polars as pl

from kagglehub.datasets import PolarsFrameType, dataset_download


# This is a thin wrapper around pl.read_database so we get the connection
# closing for free after we've produced the DataFrame
def wrapped_read_database(sql_query: str, path: str) -> pl.DataFrame:
    with sqlite3.connect(path) as conn:
        df = pl.read_database(sql_query, conn)
        return df


# These are the currently supported read_* functions for the polars adapter.
# Taken from https://docs.pola.rs/api/python/stable/reference/io.html. More can be
# added as demand dictates.
SUPPORTED_READ_FUNCTIONS_BY_EXTENSION: dict[str, Callable] = {
    ".csv": pl.read_csv,
    ".tsv": pl.read_csv,  # Additional kwargs for the polars method are defined below
    ".json": pl.read_json,
    ".jsonl": pl.read_ndjson,
    ".parquet": pl.read_parquet,
    ".feather": pl.read_ipc,
    # Technically, SQLite files can have any extension (or none), but these are the listed
    # ones on Wikipedia: https://en.wikipedia.org/wiki/SQLite. We can add more robust checking
    # by testing connections if users request it.
    ".sqlite": wrapped_read_database,
    ".sqlite3": wrapped_read_database,
    ".db": wrapped_read_database,
    ".db3": wrapped_read_database,
    ".s3db": wrapped_read_database,
    ".dl3": wrapped_read_database,
    # read_excel supports many file types:
    # https://docs.pola.rs/api/python/stable/reference/api/polars.read_excel.html
    ".xls": pl.read_excel,
    ".xlsx": pl.read_excel,
    ".xlsb": pl.read_excel,
    # The default engine for polars Excel reading is "calamine" which uses "the fastexcel module to bind the Rust-based
    # Calamine parser". This does seem to support xlsm despite polars not saying: https://github.com/tafia/calamine
    ".xlsm": pl.read_excel,
    # There's no indication that polars supports odf or odt files like pandas does, so just allow this one for now
    ".ods": pl.read_excel,
}

# These are the currently supported scan_* functions for the polars adapter.
# Taken from https://docs.pola.rs/api/python/stable/reference/io.html. These are the
# only methods that support LazyFrames (the preferred way of using polars) and are
# a subset of the more extensive read_* methods, which produce eager DataFrames.
SUPPORTED_SCAN_FUNCTIONS_BY_EXTENSION: dict[str, Callable] = {
    ".csv": pl.scan_csv,
    ".tsv": pl.scan_csv,  # Additional kwargs for the polars method are defined below
    ".jsonl": pl.scan_ndjson,
    ".parquet": pl.scan_parquet,
    ".feather": pl.scan_ipc,
}


# Certain extensions leverage a shared method but require additional static kwargs
STATIC_KWARGS_BY_EXTENSION: dict[str, dict[str, str | bool]] = {".tsv": {"separator": "\t"}}
MISSING_SQL_QUERY_ERROR_MESSAGE = "Loading from a SQLite file requires a SQL query"


def load_polars_dataset(
    handle: str,
    path: str,
    *,
    polars_frame_type: PolarsFrameType = PolarsFrameType.LAZY_FRAME,
    polars_kwargs: Any = None,  # noqa: ANN401
    sql_query: str | None,
) -> pl.DataFrame | dict[int | str, pl.DataFrame]:
    """Creates polars LazyFrame(s) or DataFrame(s) from a file in the dataset

    Args:
        handle: (string) The dataset handle
        path: (string) Path to a file within the dataset
        polars_frame_type:
            (PolarsFrameType) Optional control for which Frame to return: LazyFrame or DataFrame. The default is
            PolarsFrameType.LAZY_FRAME.

            PolarsFrameType.LAZY_FRAME: We attempt to use a scan_* method if it's available for the provided file
            extension. Otherwise, we use a read_* method to produce a DataFrame and return the result after calling
            .lazy() on it. This satisfies the requested polars_frame_type as a LazyFrame, but does require loading the
            file in memory.

            PolarsFrameType.DATA_FRAME: We use whatever read_* method corresponds to the provided file extension and
            return the resulting DatFrame.
        polars_kwargs:
            (dict) Optional set of kwargs to pass to the polars `read_*` method while constructing the DataFrame(s)
        sql_query:
            (string) Argument to be used for SQLite files. Required when reading a SQLite file. See polars documentation
            for details: https://docs.pola.rs/api/python/stable/reference/api/polars.read_database.html

    Returns:
        - dict[int | str, LazyFrame] or dict[int | str, DataFrame] for Excel-like files with multiple sheets
        - A polars LazyFrame or DataFrame for all others

    Raises:
        ValueError: If the file extension is not supported or the file fails to read
    """
    polars_kwargs = {} if polars_kwargs is None else polars_kwargs
    file_extension = os.path.splitext(path)[1]
    io_function, io_frame_type = _validate_io_function(file_extension, sql_query, polars_frame_type)

    # Now that everything has been validated, we can start downloading and processing
    filepath = dataset_download(handle, path)
    try:
        result = io_function(
            *_build_args(io_function, filepath, sql_query),
            **_build_kwargs(file_extension, polars_kwargs),
        )
    except Exception as e:
        read_error_message = f"Error reading file: {e}"
        raise ValueError(read_error_message) from e

    # The user requested a LazyFrame, but there's no scan_* method for the file extension. We need to
    # convert the resulting DataFrame(s) to a LazyFrame(s) before we return the result.
    if io_frame_type is PolarsFrameType.DATA_FRAME and polars_frame_type is PolarsFrameType.LAZY_FRAME:
        if isinstance(result, dict):
            for key, value in result.items():
                # The only time a dict can be returned is for Excel files, which only supports read_excel.
                # So all of these *should* already be DataFrames, but we'll just be extra safe.
                if isinstance(value, pl.DataFrame):
                    result[key] = value.lazy()
        else:
            result = result.lazy()

    return result


def _validate_io_function(
    file_extension: str, sql_query: str | None, polars_frame_type: PolarsFrameType
) -> tuple[Callable, PolarsFrameType]:
    if file_extension not in SUPPORTED_READ_FUNCTIONS_BY_EXTENSION:
        extension_error_message = (
            f"Unsupported file extension: '{file_extension}'. "
            f"Supported file extensions are: {', '.join(SUPPORTED_READ_FUNCTIONS_BY_EXTENSION.keys())}"
        )
        raise ValueError(extension_error_message) from None

    read_function = SUPPORTED_READ_FUNCTIONS_BY_EXTENSION[file_extension]
    if read_function is wrapped_read_database and not sql_query:
        raise ValueError(MISSING_SQL_QUERY_ERROR_MESSAGE)

    # Now that we've validated all the inputs, we can determine whether we can/should use a scan_* function
    if polars_frame_type is PolarsFrameType.DATA_FRAME or file_extension not in SUPPORTED_SCAN_FUNCTIONS_BY_EXTENSION:
        return (read_function, PolarsFrameType.DATA_FRAME)

    return (SUPPORTED_SCAN_FUNCTIONS_BY_EXTENSION[file_extension], PolarsFrameType.LAZY_FRAME)


def _build_args(read_function: Callable, path: str, sql_query: str | None) -> list:
    # The presence of the sql_query arg was already validated in _validate_read_function
    return [path] if read_function != wrapped_read_database else [sql_query, path]


def _build_kwargs(file_extension: str, polars_kwargs: Any) -> dict:  # noqa: ANN401
    static_kwargs: dict[str, Any] = (
        {} if file_extension not in STATIC_KWARGS_BY_EXTENSION else STATIC_KWARGS_BY_EXTENSION[file_extension]
    )
    # NOTE: Order matters here as we're letting users override the static args. This may be valuable in the edge case
    # that a CSV/TSV has some other separator than what the extension indicates (we see this in real datasets).
    return {**static_kwargs, **polars_kwargs}



import logging
import os
import tempfile
import time
import unittest
import uuid
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TypeVar

from kagglehub import dataset_upload, datasets_helpers
from kagglehub.clients import build_kaggle_client
from kagglehub.config import get_kaggle_credentials
from kagglehub.handle import parse_dataset_handle
from kagglehub.http_resolver import _get_current_version

logger = logging.getLogger(__name__)


ReturnType = TypeVar("ReturnType")


class TestDatasetUpload(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.mkdtemp()
        self.dummy_files = ["foo.txt", "bar.csv"]
        for file in self.dummy_files:
            with open(os.path.join(self.temp_dir, file), "w") as f:
                f.write("dummy content")
        self.dataset_uuid = str(uuid.uuid4())
        credentials = get_kaggle_credentials()
        if not credentials:
            self.fail("Make sure to set your Kaggle credentials before running the tests")
        self.owner_slug = credentials.username
        self.dataset_slug = f"dataset-{self.dataset_uuid}"
        self.handle = f"{self.owner_slug}/{self.dataset_slug}"

    def test_dataset_upload_and_versioning(self) -> None:
        # Create Dataset
        self.assertTrue(dataset_upload_and_wait(self.handle, self.temp_dir))

        # Create Version
        dataset_upload(self.handle, self.temp_dir, "new version")

        # If delete dataset does not raise an error, then the upload was successful.

    def test_dataset_upload_and_versioning_zip(self) -> None:
        with TemporaryDirectory() as temp_dir:
            for i in range(60):
                test_filepath = Path(temp_dir) / f"temp_test_file_{i}"
                test_filepath.touch()

            # Create Dataset
            self.assertTrue(dataset_upload_and_wait(self.handle, self.temp_dir))

            # Create Version
            # dataset_upload(self.handle, self.temp_dir, "new version")

    def test_dataset_upload_directory(self) -> None:
        with TemporaryDirectory() as temp_dir:
            # Create new folder within temp_dir
            inner_folder_path = Path(temp_dir) / "inner_folder"
            inner_folder_path.mkdir()

            for i in range(60):
                # Create a file in the temp_dir
                test_filepath = Path(temp_dir) / f"temp_test_file_{i}"
                test_filepath.touch()

                # Create the same file in the inner_folder
                test_filepath_inner = inner_folder_path / f"temp_test_file_{i}"
                test_filepath_inner.touch()

            # Create Dataset
            self.assertTrue(dataset_upload_and_wait(self.handle, self.temp_dir))

            # Create Version
            dataset_upload(self.handle, self.temp_dir, "new version")

    def test_dataset_upload_directory_structure(self) -> None:
        nested_dir = Path(self.temp_dir) / "nested"
        nested_dir.mkdir()

        with open(Path(self.temp_dir) / "file1.txt", "w") as f:
            f.write("dummy content in nested file")

        # Create dummy files in the nested directory
        nested_dummy_files = ["nested_dataset.txt", "nested_csv.csv"]
        for file in nested_dummy_files:
            with open(nested_dir / file, "w") as f:
                f.write("dummy content in nested file")

        # Call the dataset upload function with the base directory
        dataset_upload(self.handle, self.temp_dir)

    def test_dataset_upload_nested_dir(self) -> None:
        # Create a nested directory within self.temp_dir
        nested_dir = Path(self.temp_dir) / "nested"
        nested_dir.mkdir()

        # Create dummy files in the nested directory
        nested_dummy_files = ["nested_dataset.txt", "nested_csv.csv"]
        for file in nested_dummy_files:
            with open(nested_dir / file, "w") as f:
                f.write("dummy content in nested file")

        # Call the dataset upload function with the base directory
        dataset_upload(self.handle, self.temp_dir)

    def test_single_file_upload(self) -> None:
        single_file_path = Path(self.temp_dir) / "dummy_file.txt"
        with open(single_file_path, "wb") as f:
            f.write(os.urandom(100))

        dataset_upload(self.handle, str(single_file_path))

    def tearDown(self) -> None:
        time.sleep(5)  # hacky. Need to wait until a dataset is ready to be deleted.
        datasets_helpers.delete_dataset(self.owner_slug, self.dataset_slug)


# TODO(b/379171781): Remove waiting logic once uploading a new version while the first version is being processed is
# supported.
def dataset_upload_and_wait(handle: str, local_dataset_dir: str, version_notes: str = "") -> bool:
    dataset_upload(handle, local_dataset_dir, version_notes)
    h = parse_dataset_handle(handle)
    time.sleep(1)
    with build_kaggle_client() as api_client:
        max_attempts = 10
        for _attempt in range(max_attempts):
            try:
                if _get_current_version(api_client, h) > 0:
                    return True
            except Exception:
                # wait a bit before checking for completion again.
                time.sleep(5)

        return False



from collections.abc import Callable
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Event, Thread
from types import ModuleType
from typing import Any

from kagglehub.packages import PackageScope, _apply_context_manager_to_module
from tests.fixtures import BaseTestCase

from .utils import clear_imported_kaggle_packages


class TestPackageMultithreading(BaseTestCase):

    def tearDown(self) -> None:
        clear_imported_kaggle_packages()

    def test_multithreaded_package_sequential(self) -> None:
        # Create our fake module and do the kagglehub import logic
        module = import_fake_package_module("test", package_act)

        thread_a_entered = Event()
        thread_b_entered = Event()
        thread_a_ready_for_exit = Event()
        thread_b_ready_for_exit = Event()

        def thread_a_worker() -> None:
            module.package_act(thread_a_entered, thread_a_ready_for_exit)

        def thread_b_worker() -> None:
            module.package_act(thread_b_entered, thread_b_ready_for_exit)

        thread_a = Thread(target=thread_a_worker)
        thread_b = Thread(target=thread_b_worker)

        # Thread A starts and finishes first
        thread_a.start()
        wait_or_raise(thread_a_entered)
        thread_a_ready_for_exit.set()
        thread_a.join(timeout=3)
        # Then Thread B starts and finishes afterwards
        thread_b.start()
        wait_or_raise(thread_b_entered)
        thread_b_ready_for_exit.set()
        thread_b.join(timeout=3)

    def test_multithreaded_package_interleaving_filo(self) -> None:
        # Create our fake module and do the kagglehub import logic
        module = import_fake_package_module("test", package_act)

        thread_a_entered = Event()
        thread_b_entered = Event()
        thread_a_ready_for_exit = Event()
        thread_b_ready_for_exit = Event()

        def thread_a_worker() -> None:
            module.package_act(thread_a_entered, thread_a_ready_for_exit)

        def thread_b_worker() -> None:
            module.package_act(thread_b_entered, thread_b_ready_for_exit)

        thread_a = Thread(target=thread_a_worker)
        thread_b = Thread(target=thread_b_worker)

        # Thread A starts first
        thread_a.start()
        wait_or_raise(thread_a_entered)
        # Then Thread B starts and finishes
        thread_b.start()
        wait_or_raise(thread_b_entered)
        thread_b_ready_for_exit.set()
        thread_b.join(timeout=3)
        # Then Thread A finally finishes
        thread_a_ready_for_exit.set()
        thread_a.join(timeout=3)

    def test_multithreaded_package_interleaving_fifo(self) -> None:
        # Create our fake module and do the kagglehub import logic
        module = import_fake_package_module("test", package_act)

        thread_a_entered = Event()
        thread_b_entered = Event()
        thread_a_ready_for_exit = Event()
        thread_b_ready_for_exit = Event()

        def thread_a_worker() -> None:
            module.package_act(thread_a_entered, thread_a_ready_for_exit)

        def thread_b_worker() -> None:
            module.package_act(thread_b_entered, thread_b_ready_for_exit)

        thread_a = Thread(target=thread_a_worker)
        thread_b = Thread(target=thread_b_worker)

        # Thread A starts first
        thread_a.start()
        wait_or_raise(thread_a_entered)
        # Then Thread B starts
        thread_b.start()
        wait_or_raise(thread_b_entered)
        # Then Thread A finishes
        thread_a_ready_for_exit.set()
        thread_a.join(timeout=3)
        # Then Thread B finishes
        thread_b_ready_for_exit.set()
        thread_b.join(timeout=3)


def import_fake_package_module(module_name_suffix: str, *members: Callable[..., Any] | type[Any]) -> ModuleType:
    """Makes a fake python module with our custom context manager decorations applied.

    Creates the module with the desired function and class members, and calls _apply_context_manager_to_module
    to decorate those members with our context manager scope logic."""
    module = ModuleType(f"kagglehub_package_{module_name_suffix}")
    setattr(module, "__package_version__", "0.1.0")  # noqa: B010

    # Make a fake dir with the required kagglehub_requirements file.
    # Needs to survive until we init PackageScope which reads it.
    with TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        module.__file__ = str(temp_dir_path / "__init__.py")
        (temp_dir_path / "kagglehub_requirements.yaml").write_text("format_version: 0.1.0\ndatasources: []")

        # Link all desired members to the module
        for member in members:
            setattr(module, member.__name__, member)
            member.__module__ = module.__name__

        # Do the kagglehub side of the module import
        with PackageScope(module) as package_scope:
            _apply_context_manager_to_module(module, package_scope)

    return module


def package_act(on_enter: Event, ready_for_exit: Event) -> None:
    if not PackageScope.get():
        raise ValueError()

    on_enter.set()

    if not ready_for_exit.wait(1):
        raise TimeoutError()


def wait_or_raise(event: Event, timeout: float = 1) -> None:
    if not event.wait(timeout):
        raise TimeoutError()



import logging
from collections.abc import Callable
from http import HTTPStatus
from typing import TypeVar

import requests
from kagglesdk.kaggle_object import KaggleObject

from kagglehub.handle import CompetitionHandle, ResourceHandle
from kagglehub.logger import EXTRA_CONSOLE_BLOCK

R = TypeVar("R", bound=KaggleObject)

logger = logging.getLogger(__name__)


class CredentialError(Exception):
    pass


class KaggleEnvironmentError(Exception):
    pass


class ColabEnvironmentError(Exception):
    pass


class BackendError(Exception):
    def __init__(self, message: str, error_code: int | None = None) -> None:
        self.error_code = error_code
        super().__init__(message)


class NotFoundError(Exception):
    pass


class DataCorruptionError(Exception):
    pass


class KaggleApiHTTPError(requests.HTTPError):
    def __init__(self, message: str, response: requests.Response | None = None) -> None:
        super().__init__(message, response=response)


class ColabHTTPError(requests.HTTPError):
    def __init__(self, message: str, response: requests.Response | None = None) -> None:
        super().__init__(message, response=response)


class UnauthenticatedError(Exception):
    """Exception raised for errors in the authentication process."""

    def __init__(self, message: str = "User is not authenticated") -> None:
        super().__init__(message)


class UserCancelledError(Exception):
    pass


def handle_call(fn: Callable[[], R], resource_handle: ResourceHandle | None = None) -> R:
    """Handle errors for handler NOT returning 200 status code on failure."""
    try:
        return fn()
    except requests.HTTPError as e:
        message = str(e)
        server_error_message = ""
        try:
            server_error_message = e.strerror if e.strerror else ""
            if server_error_message:
                server_error_message = f"The server reported the following issues: {server_error_message}\n"
        except requests.exceptions.JSONDecodeError as ex:
            logger.info(f"Server payload is not json. See {ex}", extra={**EXTRA_CONSOLE_BLOCK})
        resource_url = resource_handle.to_url() if resource_handle else e.response.url
        if e.response.status_code in {HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN}:
            if isinstance(resource_handle, CompetitionHandle):
                message = (
                    f"{e.response.status_code} Client Error."
                    "\n\n"
                    f"You don't have permission to access resource at URL: {resource_url}"
                    "\nPlease make sure you are authenticated and have accepted the competition rules which"
                    f" can be found at this location: {resource_url}/rules"
                )
            else:
                message = (
                    f"{e.response.status_code} Client Error."
                    "\n\n"
                    f"You don't have permission to access resource at URL: {resource_url}. "
                    f"{server_error_message}"
                    f"Please make sure you are authenticated if you are trying to access a "
                    f"private resource or a resource requiring consent."
                )

        if e.response.status_code == HTTPStatus.NOT_FOUND:
            message = (
                f"{e.response.status_code} Client Error."
                "\n\n"
                f"Resource not found at URL: {resource_url}\n"
                f"{server_error_message}"
                "Please make sure you specified the correct resource identifiers."
            )

        # Default handling
        raise KaggleApiHTTPError(message, response=e.response) from e


def colab_raise_for_status(response: requests.Response, resource_handle: ResourceHandle | None = None) -> None:
    """
    Wrapper around `response.raise_for_status()` that provides nicer error messages
    See: https://requests.readthedocs.io/en/latest/api/#requests.Response.raise_for_status
    """
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        message = str(e)
        resource_url = resource_handle.to_url() if resource_handle else response.url

        if response.status_code in {HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN}:
            message = (
                f"{response.status_code} Client Error."
                "\n\n"
                f"You don't have permission to access resource at URL: {resource_url}"
                "\nPlease make sure you are authenticated if you are trying to access a private resource or a resource"
                " requiring consent."
            )
        # Default handling
        raise ColabHTTPError(message, response=response) from e


def handle_mutate_call(fn: Callable[[], R]) -> R:
    """
    Handle errors for handlers returning 200 status code & an error message in the body on failure.

    These are mostly create & update handlers. E.g.: https://github.com/Kaggle/kaggleazure/blob/be355335448271845fc460a6e153f2cd4b030fec/Kaggle.Services.Models/Handlers/ModelApiService/V1/DeleteModelHandler.cs#L29-L31
    """
    response = fn().to_dict()
    if not (200 <= response.get("code", 200) < 300):  # noqa: PLR2004
        error_message = response.get("message", "No error message provided")
        raise BackendError(error_message)
    elif "error" in response and response["error"] != "":
        error_code = int(response["errorCode"]) if "errorCode" in response else None
        raise BackendError(response["error"], error_code)
    return response



from flask import Flask, jsonify, request
from flask.typing import ResponseReturnValue

app = Flask(__name__)


@app.route("/", methods=["HEAD"])
def head() -> ResponseReturnValue:
    return "", 200


GOOD_CREDENTIALS_USERNAME = "lastplacelarry"
GOOD_CREDENTIALS_API_TOKEN = "some-token"


@app.errorhandler(404)
def error(e: Exception):  # noqa: ANN201
    data = {"code": "404", "error": str(e)}
    return jsonify(data), 200


@app.route("/api/v1/api.v1.DiagnosticsService/Hello", methods=["POST"])
def hello() -> ResponseReturnValue:
    auth = request.authorization
    if auth and auth.token == GOOD_CREDENTIALS_API_TOKEN:
        data = {"message": "Hello from test server!", "userName": GOOD_CREDENTIALS_USERNAME}
        return jsonify(data), 200
    else:
        return jsonify({"code": 401}), 200



import pathlib
from typing import Any

import yaml

from kagglehub.handle import (
    CompetitionHandle,
    DatasetHandle,
    ModelHandle,
    NotebookHandle,
    PackageHandle,
    ResourceHandle,
    UtilityScriptHandle,
    parse_competition_handle,
    parse_dataset_handle,
    parse_model_handle,
    parse_notebook_handle,
    parse_package_handle,
    parse_utility_script_handle,
)

# Current version of the file format written here
FORMAT_VERSION = "0.1.0"

FORMAT_VERSION_FIELD = "format_version"
DATASOURCES_FIELD = "datasources"
DATASOURCE_TYPE_FIELD = "type"
DATASOURCE_REF_FIELD = "ref"
DATASOURCE_VERSION_FIELD = "version"

HANDLE_TYPE_NAMES = {
    CompetitionHandle: "Competition",
    DatasetHandle: "Dataset",
    ModelHandle: "Model",
    NotebookHandle: "Notebook",
    UtilityScriptHandle: "UtilityScript",
    PackageHandle: "Package",
}

HANDLE_TYPE_PARSERS = {
    HANDLE_TYPE_NAMES[CompetitionHandle]: parse_competition_handle,
    HANDLE_TYPE_NAMES[DatasetHandle]: parse_dataset_handle,
    HANDLE_TYPE_NAMES[ModelHandle]: parse_model_handle,
    HANDLE_TYPE_NAMES[NotebookHandle]: parse_notebook_handle,
    HANDLE_TYPE_NAMES[UtilityScriptHandle]: parse_utility_script_handle,
    HANDLE_TYPE_NAMES[PackageHandle]: parse_package_handle,
}

# Maps requested ResourceHandle (which may include version) to version used
VersionedDatasources = dict[ResourceHandle, int | None]

# Tracks datasources accessed in the current session
_accessed_datasources: VersionedDatasources = {}


def register_datasource_access(handle: ResourceHandle, version: int | None) -> None:
    """Record that a datasource was accessed.

    Link the user-requested handle to the version retrieved."""
    _accessed_datasources[handle] = version


def get_accessed_datasources() -> VersionedDatasources:
    return _accessed_datasources.copy()


def write_file(filepath: str | pathlib.Path) -> None:
    """Write the datasources accessed during this session to a yaml file.

    Args:
        filepath: (str | pathlib.Path) Where to write the yaml file.
    """
    data = {
        FORMAT_VERSION_FIELD: FORMAT_VERSION,
        DATASOURCES_FIELD: [_serialize_datasource(h, version) for h, version in _accessed_datasources.items()],
    }

    with open(filepath, "w") as f:
        yaml.dump(data, f, sort_keys=False)


def read_file(filepath: str | pathlib.Path) -> VersionedDatasources:
    """Read a yaml file with datasource + version records.

    Args:
        filepath: (str | pathlib.Path) Path of yaml file to read from.
    Returns:
        Dictionary mapping ResourceHandle to version number, parsed from file.
    """
    with open(filepath) as f:
        data = yaml.safe_load(f)

    format_version = data.get(FORMAT_VERSION_FIELD)
    if format_version != FORMAT_VERSION:
        msg = f"Unsupported tracker file format version: {format_version}"
        raise ValueError(msg)

    versioned_datasources: VersionedDatasources = {}
    for datasource in data.get(DATASOURCES_FIELD, []):
        h, version = _deserialize_datasource(datasource)
        versioned_datasources[h] = version

    return versioned_datasources


def _serialize_datasource(h: ResourceHandle, version: int | None) -> dict:
    data: dict[str, Any] = {
        DATASOURCE_TYPE_FIELD: HANDLE_TYPE_NAMES[type(h)],
        DATASOURCE_REF_FIELD: str(h),
    }

    if version is not None:
        data[DATASOURCE_VERSION_FIELD] = version

    return data


def _deserialize_datasource(data: dict) -> tuple[ResourceHandle, int | None]:
    parser = HANDLE_TYPE_PARSERS[data[DATASOURCE_TYPE_FIELD]]
    h = parser(data[DATASOURCE_REF_FIELD])
    version = _parse_version(data.get(DATASOURCE_VERSION_FIELD, None))

    return h, version


def _parse_version(version: Any) -> int | None:  # noqa: ANN401
    if version is None or isinstance(version, int):
        return version

    if isinstance(version, str):
        try:
            return int(version)
        except:  # noqa: E722, S110
            # Fall through to the raise below
            pass

    msg = f"Invalid version: '{version}'. Expected an integer or None."
    raise ValueError(msg)



import os
from unittest import mock

import kagglehub
from kagglehub.cache import NOTEBOOKS_CACHE_SUBFOLDER, get_cached_archive_path
from kagglehub.exceptions import UserCancelledError
from kagglehub.handle import parse_package_handle
from tests.fixtures import BaseTestCase

from .server_stubs import notebook_output_download_stub as stub
from .server_stubs import serv
from .utils import clear_imported_kaggle_packages, create_test_cache, login, parameterized

INVALID_ARCHIVE_PACKAGE_HANDLE = "invalid/invalid/invalid/invalid/invalid"
VERSIONED_PACKAGE_HANDLE = "dster/package-test/versions/1"
UNVERSIONED_PACKAGE_HANDLE = "dster/package-test"

EXPECTED_NOTEBOOK_SUBDIR = os.path.join(NOTEBOOKS_CACHE_SUBFOLDER, "dster", "package-test", "output", "versions", "1")

YES_INPUTS = ["yes", "y", "YES", "Y", "YeS"]


class TestHttpPackageImport(BaseTestCase):

    def tearDown(self) -> None:
        clear_imported_kaggle_packages()

    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    @parameterized(*YES_INPUTS)
    def test_package_versioned_succeeds(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)

                mock_input.assert_called_once()
                self.assertIn("foo", dir(package))
                self.assertEqual("bar", package.foo())

                archive_path = get_cached_archive_path(parse_package_handle(VERSIONED_PACKAGE_HANDLE))
                self.assertFalse(os.path.exists(archive_path))

    def test_package_versioned_bypass_confirmation_succeeds(self) -> None:
        with create_test_cache():
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE, bypass_confirmation=True)

            self.assertIn("foo", dir(package))
            self.assertEqual("bar", package.foo())

            archive_path = get_cached_archive_path(parse_package_handle(VERSIONED_PACKAGE_HANDLE))
            self.assertFalse(os.path.exists(archive_path))

    def test_package_versioned_user_owned_succeeds(self) -> None:
        login("some-token")

        with create_test_cache():
            package = kagglehub.package_import(VERSIONED_PACKAGE_HANDLE, bypass_confirmation=True)

            self.assertIn("foo", dir(package))
            self.assertEqual("bar", package.foo())

            archive_path = get_cached_archive_path(parse_package_handle(VERSIONED_PACKAGE_HANDLE))
            self.assertFalse(os.path.exists(archive_path))

    @parameterized(*YES_INPUTS)
    def test_package_unversioned_succeeds(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                package = kagglehub.package_import(UNVERSIONED_PACKAGE_HANDLE)

                mock_input.assert_called_once()
                self.assertIn("foo", dir(package))
                self.assertEqual("baz", package.foo())

                archive_path = get_cached_archive_path(parse_package_handle(UNVERSIONED_PACKAGE_HANDLE))
                self.assertFalse(os.path.exists(archive_path))

    @parameterized("no", "NO", "n", "", "anything but yes")
    def test_package_versioned_aborts(self, input_value: str) -> None:
        with mock.patch("builtins.input", return_value=input_value) as mock_input:
            with create_test_cache():
                with self.assertRaises(UserCancelledError):
                    kagglehub.package_import(VERSIONED_PACKAGE_HANDLE)
                mock_input.assert_called_once()

    def test_notebook_download_bad_archive(self) -> None:
        with create_test_cache():
            with self.assertRaises(ValueError):
                kagglehub.package_import(INVALID_ARCHIVE_PACKAGE_HANDLE)



import contextlib
import contextvars
import functools
import importlib
import inspect
import logging
import pathlib
import re
import subprocess
import sys
import tempfile
import textwrap
from collections.abc import Callable
from types import ModuleType
from typing import Any, Optional

from kagglesdk.kaggle_env import is_in_kaggle_notebook

from kagglehub import registry
from kagglehub.auth import get_username
from kagglehub.cache import get_cached_path
from kagglehub.env import is_in_colab_notebook
from kagglehub.exceptions import UserCancelledError
from kagglehub.handle import PackageHandle, ResourceHandle, parse_package_handle
from kagglehub.logger import EXTRA_CONSOLE_BLOCK
from kagglehub.tracker import VersionedDatasources, read_file

logger = logging.getLogger(__name__)

# Current version of the package format used by Kaggle Packages
PACKAGE_VERSION = "0.1.0"
# Name of module field referring to its package version
PACKAGE_VERSION_NAME = "__package_version__"

# Name of directory within Notebook Output where the Kaggle Package lives
PACKAGE_NOTEBOOK_DIR = "package"

# Relative path within an exported package where asset files are stored
EXPORTED_PACKAGE_ASSETS_DIR = "assets"
# Absolute path for writing asset files during Package creation in Kaggle Notebooks
KAGGLE_NOTEBOOK_ASSETS_STAGING_PATH = "/kaggle/package_assets"

# Handle for the Kaggle Notebook of the Package's Dependency Manager
DEPENDENCY_MANAGER_HANDLE_NAME = "__dependency_manager_notebook__"
# Filename for Dependency Manager Notebook's install script
DEPENDENCY_MANAGER_INSTALL_FILEPATH = "install_requirements.sh"

# Expected name of the kagglehub requirements file
KAGGLEHUB_REQUIREMENTS_FILENAME = "kagglehub_requirements.yaml"


def package_import(
    handle: str, *, force_download: bool | None = False, bypass_confirmation: bool = False
) -> ModuleType:
    """Download a Kaggle Package and import it.

    A Kaggle Package is a Kaggle Notebook which has exported code to a python package format.

    Args:
        handle: (string) the notebook handle under https://kaggle.com/code.
        force_download: (bool) Optional flag to force download motebook output, even if it's cached.
    Returns:
        The imported python module.
    """
    h = parse_package_handle(handle)

    logger.info(f"Downloading Notebook Output for Package: {h.to_url()} ...", extra={**EXTRA_CONSOLE_BLOCK})
    notebook_path, version = registry.notebook_output_resolver(h, path=None, force_download=force_download)
    if not h.is_versioned() and version:
        h = h.with_version(version)

    init_file_path = pathlib.Path(notebook_path) / PACKAGE_NOTEBOOK_DIR / "__init__.py"
    if not init_file_path.exists():
        msg = f"Notebook '{h!s}' is not a Package, missing '{PACKAGE_NOTEBOOK_DIR}/__init__.py' file."
        raise ValueError(msg)

    # Unique module name based on handle + downloaded version
    module_name = re.sub(r"[^a-zA-Z0-9_]", "_", f"kagglehub_package_{h.owner}_{h.notebook}_{version}")

    # If this module was already imported and the user didn't re-download, just return it
    if module_name in sys.modules and not force_download:
        return sys.modules[module_name]

    # We're going to run the downloaded code (via `import`) so get user confirmation first.
    if not bypass_confirmation:
        _confirm_import(h)

    # If this module was already imported but now re-downloaded, clear it and any submodules before re-importing
    if module_name in sys.modules:
        logger.info(
            f"Uninstalling existing package module {module_name} before re-importing.", extra={**EXTRA_CONSOLE_BLOCK}
        )
        del sys.modules[module_name]
        submodule_names = [name for name in sys.modules if name.startswith(f"{module_name}.")]
        for name in submodule_names:
            del sys.modules[name]

    spec = importlib.util.spec_from_file_location(module_name, init_file_path)
    if spec is None or spec.loader is None:
        msg = f"Could not load module from {init_file_path} as {module_name}."
        raise ImportError(msg)

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module


def _confirm_import(h: PackageHandle) -> None:
    """Gets user's manual confirmation before importing a package which would execute arbitrary code.

    This is skipped for known safe cases:
    1. Within a Kaggle or Colab environment (already containerized, not user-owned system)
    2. Package owned by the logged-in user (assumed trusted)
    3. Nested Package (assume confirmation already provided)
    """
    if (
        is_in_kaggle_notebook()
        or is_in_colab_notebook()
        or get_username() == h.owner
        # Nested package, we assume confirmation has already been provided for the top-level Package.
        or PackageScope.get() is not None
    ):
        return

    msg = f"""
      WARNING: You are about to execute untrusted code.
      This code could modify your python environment or operating system.

      Review this code at {h.to_url()}
      or in your download cache at {get_cached_path(h)}

      It is strongly recommended that you run this code within a container
      such as Docker to provide a secure, isolated execution environment.
      See https://www.kaggle.com/docs/packages for more information.
      """

    confirmation = input(f"{textwrap.dedent(msg)}\nDo you want to proceed? (y)es/[no]: ")
    if confirmation.lower() not in ["y", "yes"]:
        raise UserCancelledError()


def get_package_asset_path(path: str) -> pathlib.Path:
    """Returns a path referring to an asset file for use in Kaggle Packages.

    If within a PackageScope context, returns the path relative to it. This should be true
    any time a Package has been imported, whether via `package_import()` or directly.
    Otherwise, assumes we're in an interactive Kaggle Notebook and creating a Package,
    where package data should be written to a staging directory which will later get copied
    into the exported package when the Notebook is saved.

    Args:
        path: (str) The relative path of the desired asset file.
    Returns:
        The absolute path (as pathlib.Path) of the desired asset file.
    """
    package_scope = PackageScope.get()
    if package_scope:
        return package_scope.path / EXPORTED_PACKAGE_ASSETS_DIR / path

    if not is_in_kaggle_notebook():
        msg = (
            "Asset paths should only be retrieved within the execution of an existing Kaggle "
            "Package, or during the creation of a new Kaggle Package within a Kaggle Notebook."
        )
        raise ValueError(msg)

    asset_path = pathlib.Path(KAGGLE_NOTEBOOK_ASSETS_STAGING_PATH) / path
    asset_path.parent.mkdir(parents=True, exist_ok=True)
    return asset_path


def _finalize_package_import(package_module: ModuleType) -> list[str]:
    """Complete the import of a Kaggle Package's python module.

    Only intended for use by auto-generated Package __init__.py files.

    Installs the package's Python Dependencies if applicable, and imports all (public) sibling .py
    files as submodules, exposes their public members onto the parent module, and decorates
    them with our PackageScope (see that class for more details).

    Args:
        package_module: (ModuleType) The python module of the Kaggle Package.
    Returns:
        The names of all public members which we scoped onto the module (similar to `__all__`).
    """
    package_version = getattr(package_module, PACKAGE_VERSION_NAME, None)
    if package_version != PACKAGE_VERSION:
        msg = f"Unsupported Kaggle Package version: {package_version}"
        raise ValueError(msg)

    # Create our PackageScope and apply it while we complete the import process
    with PackageScope(package_module) as package_scope:
        # Install python dependencies as optionally configured via Kaggle Notebook's Dependency Manager
        _install_dependencies(package_module)

        # Import all public submodules and track their public (__all__) members.
        all_names: set[str] = set()
        for filepath in sorted(package_scope.path.glob("[!_]*.py")):
            # Programmatically import the submodule
            submodule = importlib.import_module(f".{filepath.stem}", package=package_module.__name__)

            # Decorate all functions defined by the submodule to apply our PackageScope
            _apply_context_manager_to_module(submodule, package_scope)

            # Expose all public submodule members onto our parent module
            for name in submodule.__all__:
                setattr(package_module, name, getattr(submodule, name))
                all_names.add(name)

        return sorted(all_names)


def _install_dependencies(package_module: ModuleType) -> None:
    """Download and install a Package's dependencies, if applicable.

    Uses the Dependency Manager feature from Kaggle Notebooks."""
    dependency_manager_handle_str = getattr(package_module, DEPENDENCY_MANAGER_HANDLE_NAME, None)
    if not dependency_manager_handle_str:
        return

    package_name = package_module.__name__
    handle = parse_package_handle(dependency_manager_handle_str)
    if handle.owner != "packagemanager":
        msg = f"Package '{package_name}' has Dependency Manager '{handle!s}' not owned by 'packagemanager' user."
        raise ValueError(msg)

    logger.info(f"Downloading Dependency Manager Notebook: {handle.to_url()} ...", extra={**EXTRA_CONSOLE_BLOCK})
    path, _ = registry.notebook_output_resolver(handle)
    install_path = pathlib.Path(path) / DEPENDENCY_MANAGER_INSTALL_FILEPATH
    if not install_path.is_file():
        msg = f"Package '{package_name}' has dependencies configured but installer not found at '{install_path!s}'."
        raise ValueError(msg)

    # Run the install script and log the output to a tempfile
    # TODO(b/393170778): Use more platform-agnostic installation logic
    with tempfile.NamedTemporaryFile(
        prefix="kagglehub-package-dependencies-install-", suffix=".txt", delete=False, mode="w+"
    ) as log_file:
        logger.info(
            f"Installing python dependencies for Package '{package_name}', logging progress to '{log_file.name}'."
        )
        log_file.write(
            f"Installing python dependencies for Package '{package_name}' using install script at '{install_path!s}'.\n"
        )
        log_file.flush()

        subprocess.run(  # noqa: S603
            ["/bin/bash", str(install_path)], check=True, text=True, stdout=log_file, stderr=log_file
        )
        log_file.flush()


class PackageScope:
    """Captures data about a Kaggle Package. Use as Context Manager to apply the scope.

    When scope is applied, certain `kagglehub` calls will utilize the Package info.
    Specifically, downloading a datasource without a version specified will check
    the PackageScope and use the Package's version if it finds a matching entry.
    `kagglehub.get_package_asset_path` also relies on the current scope to pull
    assets related to a package.
    """

    # Global context variable tracking the currently active package scope.
    _current_scope_ctx = contextvars.ContextVar[Optional["PackageScope"]]("kagglehub_package_scope", default=None)
    # Global context variable tracking the context token stack for _current_scope_ctx to
    # revert to previous scope (if applicable) when exiting a given scope.
    # b/417707383: Use a ContextVar here to support multithreaded usage of an imported package.
    _token_stack_ctx = contextvars.ContextVar[list[contextvars.Token] | None](
        "kagglehub_package_scope_token_stack", default=None
    )

    def __init__(self, package_module: ModuleType):
        if not hasattr(package_module, "__file__") or not package_module.__file__:
            msg = f"Package module '{package_module.__name__}' missing '__file__'."
            raise ValueError(msg)

        self.package_module: ModuleType = package_module
        self.path: pathlib.Path = pathlib.Path(package_module.__file__).parent
        self.datasources: VersionedDatasources = read_file(self.path / KAGGLEHUB_REQUIREMENTS_FILENAME)

    def __enter__(self):
        token = PackageScope._current_scope_ctx.set(self)

        if PackageScope._token_stack_ctx.get() is None:
            PackageScope._token_stack_ctx.set([])
        PackageScope._token_stack_ctx.get().append(token)

        return self

    def __exit__(self, exc_type, exc_value, traceback):  # noqa: ANN001
        token = PackageScope._token_stack_ctx.get().pop()
        PackageScope._current_scope_ctx.reset(token)

    @staticmethod
    def get() -> Optional["PackageScope"]:
        """Gets the currently applied PackageScope, or None if none applied."""
        return PackageScope._current_scope_ctx.get()

    @staticmethod
    def get_version(h: ResourceHandle) -> int | None:
        """Gets version number for given resource within current PackageScope (if any).

        Returns None if no PackageScope is applied, or if it didn't contain the resource."""
        package_scope = PackageScope.get()

        return package_scope.datasources.get(h) if package_scope else None


def _apply_context_manager_to_module(module: ModuleType, context_manager: contextlib.AbstractContextManager) -> None:
    """Decorates all functions/methods in the module to apply the Context Manager.

    Also includes methods defined on classes."""

    # Our main decorator: Wrap function execution in our Context Manager.
    def decorate(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):  # noqa: ANN002, ANN003, ANN202
            with context_manager:
                return func(*args, **kwargs)

        return wrapper

    # Apply the decorator to all functions/methods defined in the module, also in nested classes.
    stack: list[Any] = [module]
    while stack:
        obj = stack.pop()
        for name, member in inspect.getmembers(obj):
            # Only decorate things defined within the module.
            if getattr(member, "__module__", None) != module.__name__:
                continue
            # Recurse on a class to decorate its functions / methods too.
            # These denylisted entries cause infinite loops.
            elif inspect.isclass(member) and name not in ["__base__", "__class__"]:
                stack.append(member)
            elif inspect.isfunction(member) or inspect.ismethod(member):
                setattr(obj, name, decorate(member))



import abc
from typing import Generic, TypeVar

from kagglehub.handle import ResourceHandle
from kagglehub.tracker import register_datasource_access

T = TypeVar("T", bound=ResourceHandle)


class Resolver(Generic[T]):
    """Resolver base class: all resolvers inherit from this class."""

    __metaclass__ = abc.ABCMeta

    def __call__(
        self, handle: T, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        """Resolves a handle into a path with the requested file(s) and the resource's version number.

        Args:
            handle: (T) the ResourceHandle to resolve.
            path: (string) Optional path to a file within the resource.
            force_download: (bool) Optional flag to force download, even if it's cached.

        Returns:
            A tuple of: (string representing the path, version number of resolved datasource if present)
            Some cases where version number might be missing: Competition datasource, API-based models.
        """
        path, version = self._resolve(handle, path, force_download=force_download)

        # Note handles are immutable, so _resolve() could not have altered our reference
        register_datasource_access(handle, version)

        return path, version

    @abc.abstractmethod
    def _resolve(
        self, handle: T, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        """Resolves a handle into a path with the requested file(s) and the resource's version number.

        Args:
            handle: (T) the ResourceHandle to resolve.
            path: (string) Optional path to a file within the resource.
            force_download: (bool) Optional flag to force download, even if it's cached.

        Returns:
            A tuple of: (string representing the path, version number of resolved datasource if present)
            Some cases where version number might be missing: Competition datasource, API-based models.
        """
        pass

    @abc.abstractmethod
    def is_supported(self, handle: T, path: str | None = None) -> bool:
        """Returns whether the current environment supports this handle/path."""
        pass



import logging
from pathlib import Path

from kagglehub.handle import ModelHandle
from kagglehub.models_helpers import signing_token

logger = logging.getLogger(__name__)


def sign_with_sigstore(local_model_dir: str, handle: ModelHandle) -> bool:
    from model_signing.signing import Config  # noqa: PLC0415

    try:
        token = signing_token(handle.owner, handle.model)
        if token:
            signing_file = Path(local_model_dir) / ".kaggle" / "signing.json"
            signing_file.parent.mkdir(exist_ok=True, parents=True)
            signing_file.unlink(missing_ok=True)
            # The below will throw an exception if the token can't be verified (Needs to be a production token)
            # Setting KAGGLE_API_ENDPOINT to localhost will throw the exception as stated above.
            Config().use_sigstore_signer(identity_token=token, use_staging=False).sign(
                Path(local_model_dir), signing_file
            )
            return True
        else:
            # skips transparency log publishing as we are unable to get a token
            logger.warning("Unable to retrieve identity token. Skipping signing...")
            return False
    except Exception:
        logger.exception("Signing failed. Skipping signing...")
        return False



import logging

from kagglesdk.blobs.types.blob_api_service import ApiBlobType

from kagglehub import registry
from kagglehub.gcs_upload import normalize_patterns, upload_files_and_directories
from kagglehub.handle import parse_model_handle
from kagglehub.logger import EXTRA_CONSOLE_BLOCK
from kagglehub.models_helpers import create_model_if_missing, create_model_instance_or_version
from kagglehub.signing import sign_with_sigstore

logger = logging.getLogger(__name__)

# Patterns that are always ignored for model uploading.
DEFAULT_IGNORE_PATTERNS = [".git/", "*/.git/", ".cache/", ".huggingface/"]


def model_download(
    handle: str,
    path: str | None = None,
    *,
    force_download: bool | None = False,
) -> str:
    """Download model files.

    Args:
        handle: (string) the model handle.
        path: (string) Optional path to a file within the model bundle.
        force_download: (bool) Optional flag to force download a model, even if it's cached.

    Returns:
        A string representing the path to the requested model files.
    """
    h = parse_model_handle(handle)
    logger.info(f"Downloading Model: {h.to_url()} ...", extra={**EXTRA_CONSOLE_BLOCK})
    path, _ = registry.model_resolver(h, path, force_download=force_download)
    return path


def model_upload(
    handle: str,
    local_model_dir: str,
    license_name: str | None = None,
    version_notes: str = "",
    ignore_patterns: list[str] | str | None = None,
    *,
    sigstore: bool | None = False,
) -> None:
    """Upload model files.

    Args:
        handle: (str) the model handle.
        local_model_dir: (str) path to a file in a local directory.
        license_name: (str) model license.
        version_notes: (str, optional) model versions.
        ignore_patterns (str or list[str], optional):
            Additional ignore patterns to DEFAULT_IGNORE_PATTERNS.
            Files matching any of the patterns are not uploaded.
            Patterns are standard wildcards that can be matched by
            https://docs.python.org/3/library/fnmatch.html.
            Use a pattern ending with "/" to ignore the whole dir,
            e.g., ".git/" is equivalent to ".git/*".
        sigstore: (bool, optional)
            Creates a trasparent ledger on sigstore. User must be an admin/editor of the model.
    """
    # parse slug
    h = parse_model_handle(handle)
    logger.info(f"Uploading Model {h.to_url()} ...")
    if h.is_versioned():
        is_versioned_exception = "The model handle should not include the version"
        raise ValueError(is_versioned_exception)

    # Create the model if it doesn't already exist
    create_model_if_missing(h.owner, h.model)

    # Model can be non-existent. Get token after model creation so signing token can be authorized.
    try:
        if sigstore:
            sigstore = sign_with_sigstore(local_model_dir, h)
    except ImportError:
        import_warning_message = (
            "The 'model_upload(...sign=True)' function requires the 'kagglehub[signing]' extras. "
            "Install them with 'pip install kagglehub[signing]'"
        )
        # Inform the user if we detect that they didn't install everything
        raise ImportError(import_warning_message) from None

    # Upload the model files to GCS
    tokens = upload_files_and_directories(
        local_model_dir,
        item_type=ApiBlobType.MODEL,
        ignore_patterns=normalize_patterns(default=DEFAULT_IGNORE_PATTERNS, additional=ignore_patterns),
    )

    create_model_instance_or_version(h, tokens, license_name, version_notes, sigstore=sigstore)



from pathlib import Path
from tempfile import TemporaryDirectory

import kagglehub
from kagglehub.handle import parse_model_handle
from kagglehub.tracker import read_file, write_file
from tests.fixtures import BaseTestCase

from .server_stubs import model_download_stub as stub
from .server_stubs import serv
from .utils import create_test_cache

UNVERSIONED_DATASET_HANDLE = "sarahjeffreson/featured-spotify-artiststracks-with-metadata"
VERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b/3"
UNVERSIONED_MODEL_HANDLE = "metaresearch/llama-2/pyTorch/13b"


class TestHttpRequirements(BaseTestCase):

    def setUp(self) -> None:
        # Clear out our tracking between tests
        kagglehub.tracker._accessed_datasources = {}

    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def test_two_models(self) -> None:
        with create_test_cache():
            kagglehub.model_download(UNVERSIONED_MODEL_HANDLE)
            kagglehub.model_download(VERSIONED_MODEL_HANDLE)

            with TemporaryDirectory() as d:
                requirements_path = str(Path(d) / "requirements.yaml")
                write_file(requirements_path)
                datasources = read_file(requirements_path)

            self.assertEqual(2, len(datasources))
            # Check the versions of each accessed datasource
            self.assertEqual(3, datasources[parse_model_handle(UNVERSIONED_MODEL_HANDLE)])
            self.assertEqual(3, datasources[parse_model_handle(VERSIONED_MODEL_HANDLE)])

    def test_no_datasources(self) -> None:
        with create_test_cache():
            with TemporaryDirectory() as d:
                requirements_path = str(Path(d) / "requirements.yaml")
                write_file(requirements_path)
                datasources = read_file(requirements_path)

            self.assertEqual(0, len(datasources))



__version__ = "0.4.1"

import kagglehub.logger  # configures the library logger.
from kagglehub import colab_cache_resolver, http_resolver, kaggle_cache_resolver, registry
from kagglehub.auth import login, whoami
from kagglehub.competition import competition_download
from kagglehub.datasets import (
    KaggleDatasetAdapter,
    PolarsFrameType,
    dataset_download,
    dataset_load,
    dataset_upload,
    load_dataset,
)
from kagglehub.models import model_download, model_upload
from kagglehub.notebooks import notebook_output_download
from kagglehub.packages import get_package_asset_path, package_import
from kagglehub.utility_scripts import utility_script_install

registry.model_resolver.add_implementation(http_resolver.ModelHttpResolver())
registry.model_resolver.add_implementation(kaggle_cache_resolver.ModelKaggleCacheResolver())
registry.model_resolver.add_implementation(colab_cache_resolver.ModelColabCacheResolver())

registry.dataset_resolver.add_implementation(http_resolver.DatasetHttpResolver())
registry.dataset_resolver.add_implementation(kaggle_cache_resolver.DatasetKaggleCacheResolver())
registry.dataset_resolver.add_implementation(colab_cache_resolver.DatasetColabCacheResolver())

registry.competition_resolver.add_implementation(http_resolver.CompetitionHttpResolver())
registry.competition_resolver.add_implementation(kaggle_cache_resolver.CompetitionKaggleCacheResolver())

registry.notebook_output_resolver.add_implementation(http_resolver.NotebookOutputHttpResolver())
registry.notebook_output_resolver.add_implementation(kaggle_cache_resolver.NotebookOutputKaggleCacheResolver())



"""Functions to convert string to kagglesdk enum values."""

import enum
import re
from typing import TypeVar

T = TypeVar("T")


# TODO(b/461859420): Consider moving to kagglesdk
def to_enum(enum_class: type[T], enum_str: str) -> T:
    enum_key = _camel_to_snake(enum_str).upper()

    try:
        return getattr(enum_class, enum_key)
    except AttributeError:
        try:
            prefix = _camel_to_snake(enum_class.__name__).upper()
            full_name = f"{prefix}_{enum_key}"
            return getattr(enum_class, full_name)
        except AttributeError:
            # Handle PY_TORCH vs PYTORCH, etc.
            full_name = full_name.replace("_", "")
            for item in vars(enum_class):
                if item.replace("_", "") == full_name:
                    return getattr(enum_class, item)
            msg = f"'{enum_str}' is not a valid ModelFramework"
            raise ValueError(msg) from None


def enum_to_str(enum: enum.Enum) -> str:
    names = str(enum).split(".")
    enum_class_name = names[0]
    enum_value_name = names[1]
    snakecase_prefix = _camel_to_snake(enum_class_name).upper()
    enum_value_name = enum_value_name.removeprefix(snakecase_prefix + "_")
    return _snake_to_camel(enum_value_name)


def _camel_to_snake(value: str) -> str:
    value = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", value)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", value).lower()


def _snake_to_camel(value: str) -> str:
    camel_str = "".join(word.capitalize() for word in value.lower().split("_"))
    # we lower only the first character
    return value[0].lower() + camel_str[1:]



import io
import logging
import os
from typing import Any

import polars as pl

from kagglehub.datasets import KaggleDatasetAdapter, PolarsFrameType, dataset_load, logger
from tests.fixtures import BaseTestCase

from .server_stubs import dataset_download_stub as stub
from .server_stubs import serv
from .utils import AUTO_COMPRESSED_FILE_NAME, create_test_cache

DATASET_HANDLE = "lastplacelarry/fake-dataset"
TEST_SPLIT_SIZE = 1
TRAIN_SPLIT_SIZE = 2
INVALID_KWARG = "this_is_not_a_kwarg"
TEXT_FILE = "foo.txt"
EXCEL_FILE = "my-spreadsheet.xlsx"
SHAPES_COLUMNS = ["shape", "degrees", "sides", "color", "date"]
SHAPES_ROW_COUNT = 3
SHAPES_COLUMNS_SUBSET = ["shape", "sides", "date"]


class TestLoadHfDataset(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def _load_hf_dataset_with_invalid_file_type_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.HUGGING_FACE,
                DATASET_HANDLE,
                TEXT_FILE,
            )
        self.assertIn(f"Unsupported file extension: '{os.path.splitext(TEXT_FILE)[1]}'", str(cm.exception))

    def _load_hf_dataset_with_other_loader_kwargs_and_assert_warning(self) -> None:
        output_stream = io.StringIO()
        handler = logging.StreamHandler(output_stream)
        logger.addHandler(handler)
        dataset_load(
            KaggleDatasetAdapter.HUGGING_FACE,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            polars_frame_type=PolarsFrameType.LAZY_FRAME,
            polars_kwargs={},
        )
        captured_output = output_stream.getvalue()
        self.assertIn(
            "polars_frame_type, polars_kwargs are invalid for KaggleDatasetAdapter.HUGGING_FACE", captured_output
        )

    def _load_hf_dataset_with_multiple_tables_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.HUGGING_FACE, DATASET_HANDLE, EXCEL_FILE, pandas_kwargs={"sheet_name": None}
            )
        self.assertIn(
            "Loading a Huggingface dataset requires the production of exactly one DataFrame", str(cm.exception)
        )

    def _load_hf_dataset_and_assert_loaded(self) -> None:
        hf_dataset = dataset_load(KaggleDatasetAdapter.HUGGING_FACE, DATASET_HANDLE, AUTO_COMPRESSED_FILE_NAME)
        self.assertEqual(SHAPES_ROW_COUNT, hf_dataset.num_rows)
        self.assertEqual(SHAPES_COLUMNS, hf_dataset.column_names)
        self.assertIn("hugging_face_data_loader", stub.shared_data.last_download_user_agent)

    def _load_hf_dataset_with_valid_kwargs_and_assert_loaded(self) -> None:
        hf_dataset = dataset_load(
            KaggleDatasetAdapter.HUGGING_FACE,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            hf_kwargs={"preserve_index": True},
        )
        self.assertEqual(SHAPES_ROW_COUNT, hf_dataset.num_rows)
        self.assertEqual([*SHAPES_COLUMNS, "__index_level_0__"], hf_dataset.column_names)

    def _load_hf_dataset_with_invalid_kwargs_and_assert_raises(self) -> None:
        with self.assertRaises(TypeError) as cm:
            dataset_load(
                KaggleDatasetAdapter.HUGGING_FACE,
                DATASET_HANDLE,
                AUTO_COMPRESSED_FILE_NAME,
                hf_kwargs={INVALID_KWARG: 777},
            )
        self.assertIn(INVALID_KWARG, str(cm.exception))

    def _load_hf_dataset_with_splits_and_assert_loaded(self) -> None:
        hf_dataset = dataset_load(KaggleDatasetAdapter.HUGGING_FACE, DATASET_HANDLE, AUTO_COMPRESSED_FILE_NAME)
        dataset_splits = hf_dataset.train_test_split(
            test_size=TEST_SPLIT_SIZE, train_size=TRAIN_SPLIT_SIZE, shuffle=False
        )
        for split_name, dataset in dataset_splits.items():
            self.assertEqual(TEST_SPLIT_SIZE if split_name == "test" else TRAIN_SPLIT_SIZE, dataset.num_rows)
            self.assertEqual(SHAPES_COLUMNS, dataset.column_names)

    def test_hf_dataset_with_other_loader_kwargs_prints_warning(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_other_loader_kwargs_and_assert_warning()

    def test_hf_dataset_with_invalid_file_type_raises(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_invalid_file_type_and_assert_raises()

    def test_hf_dataset_with_multiple_tables_raises(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_multiple_tables_and_assert_raises()

    def test_hf_dataset_succeeds(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_and_assert_loaded()

    def test_hf_dataset_with_valid_kwargs_succeeds(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_valid_kwargs_and_assert_loaded()

    def test_hf_dataset_with_invalid_kwargs_raises(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_invalid_kwargs_and_assert_raises()

    def test_hf_dataset_with_splits_succeeds(self) -> None:
        with create_test_cache():
            self._load_hf_dataset_with_splits_and_assert_loaded()


class TestLoadPandasDataset(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def _load_pandas_dataset_with_invalid_file_type_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.PANDAS,
                DATASET_HANDLE,
                TEXT_FILE,
            )
        self.assertIn(f"Unsupported file extension: '{os.path.splitext(TEXT_FILE)[1]}'", str(cm.exception))

    def _load_pandas_dataset_with_other_loader_kwargs_and_assert_warning(self) -> None:
        output_stream = io.StringIO()
        handler = logging.StreamHandler(output_stream)
        logger.addHandler(handler)
        dataset_load(
            KaggleDatasetAdapter.PANDAS,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            hf_kwargs={},
            polars_frame_type=PolarsFrameType.LAZY_FRAME,
            polars_kwargs={},
        )
        captured_output = output_stream.getvalue()
        self.assertIn(
            "hf_kwargs, polars_frame_type, polars_kwargs are invalid for KaggleDatasetAdapter.PANDAS", captured_output
        )

    def _load_pandas_simple_dataset_and_assert_loaded(
        self,
        file_extension: str,
        pandas_kwargs: Any = None,  # noqa: ANN401
    ) -> None:
        df = dataset_load(
            KaggleDatasetAdapter.PANDAS, DATASET_HANDLE, f"shapes.{file_extension}", pandas_kwargs=pandas_kwargs
        )
        self.assertEqual(SHAPES_ROW_COUNT, len(df))
        self.assertEqual(SHAPES_COLUMNS, list(df.columns))

    def _load_pandas_sqlite_dataset_and_assert_loaded(self) -> None:
        df = dataset_load(KaggleDatasetAdapter.PANDAS, DATASET_HANDLE, "shapes.db", sql_query="SELECT * FROM shapes")
        self.assertEqual(SHAPES_ROW_COUNT, len(df))
        self.assertEqual(SHAPES_COLUMNS, list(df.columns))

    def _load_pandas_dataset_with_multiple_tables_and_assert_loaded(self) -> None:
        result = dataset_load(
            KaggleDatasetAdapter.PANDAS, DATASET_HANDLE, EXCEL_FILE, pandas_kwargs={"sheet_name": None}
        )
        self.assertIsInstance(result, dict)
        self.assertEqual(["Cars", "Animals"], list(result.keys()))

    def _load_pandas_dataset_with_valid_kwargs_and_assert_loaded(self) -> None:
        expected_columns = ["degrees"]
        df = dataset_load(
            KaggleDatasetAdapter.PANDAS,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            pandas_kwargs={"usecols": expected_columns},
        )
        self.assertEqual(SHAPES_ROW_COUNT, len(df))
        self.assertEqual(expected_columns, list(df.columns))
        self.assertIn("pandas_data_loader", stub.shared_data.last_download_user_agent)

    def _load_pandas_dataset_with_invalid_kwargs_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.PANDAS,
                DATASET_HANDLE,
                AUTO_COMPRESSED_FILE_NAME,
                pandas_kwargs={INVALID_KWARG: 777},
            )
        self.assertIn(INVALID_KWARG, str(cm.exception))

    def test_pandas_dataset_with_invalid_file_type_raises(self) -> None:
        with create_test_cache():
            self._load_pandas_dataset_with_invalid_file_type_and_assert_raises()

    def test_pandas_dataset_with_other_loader_kwargs_prints_warning(self) -> None:
        with create_test_cache():
            self._load_pandas_dataset_with_other_loader_kwargs_and_assert_warning()

    def test_pandas_dataset_with_multiple_tables_succeeds(self) -> None:
        with create_test_cache():
            self._load_pandas_dataset_with_multiple_tables_and_assert_loaded()

    def test_pandas_dataset_with_valid_kwargs_succeeds(self) -> None:
        with create_test_cache():
            self._load_pandas_dataset_with_valid_kwargs_and_assert_loaded()

    def test_pandas_dataset_with_invalid_kwargs_raises(self) -> None:
        with create_test_cache():
            self._load_pandas_dataset_with_invalid_kwargs_and_assert_raises()

    def test_pandas_simple_dataset_succeeds(self) -> None:
        # This would be better as a parameterized test, but that doesn't work for subclasses:
        # https://docs.pytest.org/en/stable/how-to/unittest.html#pytest-features-in-unittest-testcase-subclasses
        test_cases = [
            ("csv", {}),
            ("tsv", {}),
            ("xml", {"parser": "etree"}),  # The default parser requires a pip install
            ("parquet", {}),
            ("feather", {}),
            ("json", {}),
            ("jsonl", {}),
        ]
        for test_case in test_cases:
            with create_test_cache():
                self._load_pandas_simple_dataset_and_assert_loaded(test_case[0], test_case[1])

    def test_pandas_sqlite_dataset_succeeds(self) -> None:
        with create_test_cache():
            self._load_pandas_sqlite_dataset_and_assert_loaded()


class TestLoadPolarsDataset(BaseTestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = serv.start_server(stub.app)

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()

    def _load_polars_dataset_with_invalid_file_type_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.POLARS,
                DATASET_HANDLE,
                TEXT_FILE,
            )
        self.assertIn(f"Unsupported file extension: '{os.path.splitext(TEXT_FILE)[1]}'", str(cm.exception))

    def _load_polars_dataset_with_other_loader_kwargs_and_assert_warning(self) -> None:
        output_stream = io.StringIO()
        handler = logging.StreamHandler(output_stream)
        logger.addHandler(handler)
        dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            pandas_kwargs={},
            hf_kwargs={},
        )
        captured_output = output_stream.getvalue()
        self.assertIn("pandas_kwargs, hf_kwargs are invalid for KaggleDatasetAdapter.POLARS", captured_output)

    def _load_polars_simple_dataset_and_assert_loaded(
        self,
        file_extension: str,
        polars_frame_type: PolarsFrameType,
    ) -> None:
        polars_frame = dataset_load(
            KaggleDatasetAdapter.POLARS, DATASET_HANDLE, f"shapes.{file_extension}", polars_frame_type=polars_frame_type
        )
        if polars_frame_type is PolarsFrameType.LAZY_FRAME:
            self.assertIsInstance(polars_frame, pl.LazyFrame)
            # Materialize the results to check rows and columns
            polars_frame = polars_frame.collect()
        else:
            self.assertIsInstance(polars_frame, pl.DataFrame)
        self.assertEqual(SHAPES_ROW_COUNT, len(polars_frame))
        self.assertEqual(SHAPES_COLUMNS, list(polars_frame.columns))

    def _load_polars_sqlite_dataset_and_assert_loaded(self) -> None:
        df = dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            "shapes.db",
            polars_frame_type=PolarsFrameType.DATA_FRAME,
            sql_query="SELECT * FROM shapes",
        )
        self.assertEqual(SHAPES_ROW_COUNT, len(df))
        self.assertEqual(SHAPES_COLUMNS, list(df.columns))

    def _load_polars_columns_subset_and_assert_loaded(self) -> None:
        df = dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            "shapes.csv",
            polars_frame_type=PolarsFrameType.DATA_FRAME,
            polars_kwargs={"columns": SHAPES_COLUMNS_SUBSET},
        )
        lf = dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            "shapes.csv",
        )
        self.assertEqual(str(df), str(lf.select(SHAPES_COLUMNS_SUBSET).collect()))

    def _load_polars_dataset_with_multiple_tables_and_assert_loaded(self, polars_frame_type: PolarsFrameType) -> None:
        result = dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            EXCEL_FILE,
            polars_frame_type=polars_frame_type,
            # sheet_id of 0 returns all sheets. This differs from pandas where sheet_name of None returns all
            polars_kwargs={"sheet_id": 0},
        )

        expected_type = pl.LazyFrame if polars_frame_type is PolarsFrameType.LAZY_FRAME else pl.DataFrame
        self.assertTrue(all(isinstance(v, expected_type) for v in result.values()))
        self.assertEqual(["Cars", "Animals"], list(result.keys()))

    def _load_polars_dataset_with_valid_kwargs_and_assert_loaded(self) -> None:
        expected_columns = ["degrees"]
        df = dataset_load(
            KaggleDatasetAdapter.POLARS,
            DATASET_HANDLE,
            AUTO_COMPRESSED_FILE_NAME,
            polars_frame_type=PolarsFrameType.DATA_FRAME,
            polars_kwargs={"columns": expected_columns},
        )
        self.assertEqual(SHAPES_ROW_COUNT, len(df))
        self.assertEqual(expected_columns, list(df.columns))
        self.assertIn("polars_data_loader", stub.shared_data.last_download_user_agent)

    def _load_polars_dataset_with_invalid_kwargs_and_assert_raises(self) -> None:
        with self.assertRaises(ValueError) as cm:
            dataset_load(
                KaggleDatasetAdapter.POLARS,
                DATASET_HANDLE,
                AUTO_COMPRESSED_FILE_NAME,
                polars_kwargs={INVALID_KWARG: 777},
            )
        self.assertIn(INVALID_KWARG, str(cm.exception))

    def test_polars_dataset_with_invalid_file_type_raises(self) -> None:
        with create_test_cache():
            self._load_polars_dataset_with_invalid_file_type_and_assert_raises()

    def test_polars_dataset_with_other_loader_kwargs_prints_warning(self) -> None:
        with create_test_cache():
            self._load_polars_dataset_with_other_loader_kwargs_and_assert_warning()

    def test_polars_dataset_with_multiple_tables_succeeds(self) -> None:
        with create_test_cache():
            self._load_polars_dataset_with_multiple_tables_and_assert_loaded(PolarsFrameType.LAZY_FRAME)
            self._load_polars_dataset_with_multiple_tables_and_assert_loaded(PolarsFrameType.DATA_FRAME)

    def test_polars_dataset_with_valid_kwargs_succeeds(self) -> None:
        with create_test_cache():
            self._load_polars_dataset_with_valid_kwargs_and_assert_loaded()

    def test_polars_dataset_with_invalid_kwargs_raises(self) -> None:
        with create_test_cache():
            self._load_polars_dataset_with_invalid_kwargs_and_assert_raises()

    def test_polars_simple_dataset_succeeds(self) -> None:
        # This would be better as a parameterized test, but that doesn't work for subclasses:
        # https://docs.pytest.org/en/stable/how-to/unittest.html#pytest-features-in-unittest-testcase-subclasses
        test_cases = [
            "csv",
            "tsv",
            "parquet",
            "feather",
            "json",
            "jsonl",
        ]
        for test_case in test_cases:
            with create_test_cache():
                self._load_polars_simple_dataset_and_assert_loaded(test_case, PolarsFrameType.LAZY_FRAME)
                self._load_polars_simple_dataset_and_assert_loaded(test_case, PolarsFrameType.DATA_FRAME)

    def test_polars_sqlite_dataset_succeeds(self) -> None:
        with create_test_cache():
            self._load_polars_sqlite_dataset_and_assert_loaded()

    def test_polars_columns_subset_succeeds(self) -> None:
        with create_test_cache():
            self._load_polars_columns_subset_and_assert_loaded()



import hashlib
import os
import re

from flask import Flask, Response, jsonify, make_response, request
from flask.typing import ResponseReturnValue
from kagglesdk.datasets.types.dataset_api_service import ApiDownloadDatasetRequest

from kagglehub.integrity import to_b64_digest
from tests.utils import get_test_file_path

app = Flask(__name__)


@app.route("/", methods=["HEAD"])
def head() -> ResponseReturnValue:
    return "", 200


@app.errorhandler(404)
def error(e: Exception):  # noqa: ANN201
    data = {"code": "404", "error": str(e), "message": "server side error"}
    return jsonify(data), 404


# We've arbitrarily chosen to test with the DownloadDataset API. Any resources would've have worked.
@app.route("/api/v1/datasets.DatasetApiService/DownloadDataset", methods=["POST"])
def dataset_download() -> ResponseReturnValue:
    r = ApiDownloadDatasetRequest.from_dict(request.get_json())
    if r.dataset_slug == "content_type_mismatch":
        html = """
        <html>
        <head>
            <title>Test</title>
        </head>
        <body>
            <h1>This is HTML content</h1>
        </body>
        </html>
        """
        resp = make_response(html)
        resp.headers["Content-Type"] = "application/json"
        resp.status_code = 404
        return resp
    if r.dataset_slug == "no-integrity":
        test_file_path = get_test_file_path("foo.txt")
        with open(test_file_path, "rb") as f:
            resp = Response(f.read())
            resp.content_type = "application/octet-stream"
            resp.content_length = os.path.getsize(test_file_path)
            return resp, 200
    if r.dataset_slug == "corrupted":
        test_file_path = get_test_file_path("foo.txt")
        with open(test_file_path, "rb") as f:
            resp = Response(f.read())
            resp.content_type = "application/octet-stream"
            resp.content_length = os.path.getsize(test_file_path)
            resp.headers["x-goog-hash"] = "md5=badhash"
            return resp, 200
    if r.dataset_slug == "good":
        test_file_path = get_test_file_path("foo.txt")
        with open(test_file_path, "rb") as f:
            start = 0
            f.seek(start)
            content = f.read()
            file_hash = hashlib.md5()
            file_hash.update(content)
            return (
                Response(
                    content,
                    headers={
                        "Content-type": "application/octet-stream",
                        "Content-Length": str(os.path.getsize(test_file_path)),
                        "Accept-Ranges": "bytes",
                        "x-goog-hash": f"crc32c=n03x6A==, md5={to_b64_digest(file_hash)}",
                    },
                ),
                200,
            )
    else:
        msg = "Unknown dataset"
        raise ValueError(msg)


@app.route("/api/v1/datasets.DatasetApiService/DownloadDataset", methods=["GET"])
def dataset_range_download() -> ResponseReturnValue:
    test_file_path = get_test_file_path("foo.txt")
    with open(test_file_path, "rb") as f:
        start = 0
        if "Range" not in request.headers:
            msg = "Expecting a range request"
            raise ValueError(msg)

        bytes_range = request.headers["Range"]
        # See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests
        # Format is bytes=<start>-<end>
        # If <end> is missing, then, it means until the end.
        # This test server only supports bytes=<start>- format since it is the only one we need.
        m = re.match("^bytes=([0-9]+)-$", bytes_range)
        if not m:
            return "", 400
        start = int(m.group(1))
        f.seek(start)
        content = f.read()
        file_hash = hashlib.md5()
        file_hash.update(content)
        return (
            Response(
                content,
                headers={
                    "Content-type": "application/octet-stream",
                    "Content-Length": str(os.path.getsize(test_file_path)),
                    "Accept-Ranges": "bytes",
                    "x-goog-hash": f"crc32c=n03x6A==, md5={to_b64_digest(file_hash)}",
                },
            ),
            200,
        )



import fnmatch
import logging
import os
import pathlib
import time
import zipfile
from collections.abc import Iterable, Sequence
from datetime import datetime
from tempfile import TemporaryDirectory

import requests
from kagglesdk.blobs.types.blob_api_service import ApiBlobType, ApiStartBlobUploadRequest
from kagglesdk.datasets.types.dataset_api_service import ApiDatasetNewFile, ApiUploadDirectoryInfo
from requests.exceptions import Timeout
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from kagglehub.clients import build_kaggle_client
from kagglehub.exceptions import BackendError, handle_call

logger = logging.getLogger(__name__)

MAX_FILES_TO_UPLOAD = 50
TEMP_ARCHIVE_FILE = "archive.zip"
MAX_RETRIES = 5
REQUEST_TIMEOUT = 600


class UploadDirectoryInfo:
    def __init__(
        self,
        name: str,
        files: list[str] | None = None,
        directories: list["UploadDirectoryInfo"] | None = None,
    ):
        self.name = name
        self.files = files if files is not None else []
        self.directories = directories if directories is not None else []

    def to_proto(self) -> ApiUploadDirectoryInfo:
        d = ApiUploadDirectoryInfo()
        d.name = self.name
        d.files = []
        for file in self.files:
            f = ApiDatasetNewFile()
            f.token = file
            d.files.append(f)
        d.directories = [d.to_proto() for d in self.directories]

        return d


def parse_datetime_string(string: str) -> datetime | str:
    time_formats = ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S.%f", "%Y-%m-%dT%H:%M:%S.%fZ"]
    for t in time_formats:
        try:
            return datetime.strptime(string[:26], t).replace(microsecond=0)  # noqa: DTZ007
        except:  # noqa: E722, S110
            pass
    return string


class File(object):  # noqa: UP004
    def __init__(self, init_dict: dict) -> None:
        parsed_dict = {k: parse_datetime_string(v) for k, v in init_dict.items()}
        self.__dict__.update(parsed_dict)

    @staticmethod
    def get_size(size: float, precision: int = 0) -> str:
        suffixes = ["B", "KB", "MB", "GB", "TB"]
        suffix_index = 0
        while size >= 1024 and suffix_index < 4:  # noqa: PLR2004
            suffix_index += 1
            size /= 1024.0
        return f"{size:.{precision}f}{suffixes[suffix_index]}"


def filtered_walk(*, base_dir: str, ignore_patterns: Sequence[str]) -> Iterable[tuple[str, list[str], list[str]]]:
    """An `os.walk` like directory tree generator with filtering.

    This method filters out files matching any ignore pattern.

    Args:
        base_dir (str): The base dir to walk in.
        ignore_patterns (Sequence[str]):
            The patterns for ignored files. These are standard wildcards relative to base_dir.

    Yields:
        Iterable[tuple[str, list[str], list[str]]]: (base_dir_path, list[dir_names], list[filtered_file_names])
    """
    for dir_path, dir_names, file_names in os.walk(base_dir):
        dir_p = pathlib.Path(dir_path)
        filtered_files = []
        for file_name in file_names:
            rel_file_p = (dir_p / file_name).relative_to(base_dir)
            if not any(fnmatch.fnmatch(name=str(rel_file_p), pat=pat) for pat in ignore_patterns):
                filtered_files.append(file_name)
        if filtered_files:
            yield (dir_path, dir_names, filtered_files)


def _check_uploaded_size(session_uri: str, file_size: int, backoff_factor: int = 1) -> int:
    """Check the status of the resumable upload."""
    headers = {"Content-Length": "0", "Content-Range": f"bytes */{file_size}"}
    retry_count = 0

    while retry_count < MAX_RETRIES:
        try:
            response = requests.put(session_uri, headers=headers, timeout=REQUEST_TIMEOUT)
            if response.status_code == 308:  # Resume Incomplete # noqa: PLR2004
                range_header = response.headers.get("Range")
                if range_header:
                    bytes_uploaded = int(range_header.split("-")[1]) + 1
                    return bytes_uploaded
                return 0  # If no Range header, assume no bytes were uploaded
            else:
                return file_size
        except (requests.ConnectionError, Timeout):
            logger.info(f"Network issue while checking uploaded size, retrying in {backoff_factor} seconds...")
            time.sleep(backoff_factor)
            backoff_factor = min(backoff_factor * 2, 60)
            retry_count += 1

    return 0  # Return 0 if all retries fail


def _upload_blob(file_path: str, item_type: ApiBlobType) -> str:
    """Uploads a file to a remote server as a blob and returns an upload token.

    Args:
        file_path: The path to the file to be uploaded.
        item_type : The type of the item associated with the file.

    Returns:
        A str token of uploaded blob.
    """
    file_size = os.path.getsize(file_path)
    with build_kaggle_client() as api_client:
        r = ApiStartBlobUploadRequest()
        r.type = item_type
        r.name = os.path.basename(file_path)
        r.content_length = file_size
        r.last_modified_epoch_seconds = int(os.path.getmtime(file_path))
        response = handle_call(lambda: api_client.blobs.blob_api_client.start_blob_upload(r))

    session_uri = response.create_url
    headers = {"Content-Type": "application/octet-stream"}

    retry_count = 0
    uploaded_bytes = 0
    backoff_factor = 1  # Initial backoff duration in seconds

    with open(file_path, "rb") as f, tqdm(total=file_size, desc="Uploading", unit="B", unit_scale=True) as pbar:
        while retry_count < MAX_RETRIES and (file_size == 0 or uploaded_bytes < file_size):
            try:
                # Special case for empty files.
                if file_size == 0:
                    headers["Content-Length"] = "0"
                    upload_data = None
                # Resumable upload for non-empty files.
                else:
                    f.seek(uploaded_bytes)
                    headers["Content-Range"] = f"bytes {uploaded_bytes}-{file_size - 1}/{file_size}"
                    upload_data = CallbackIOWrapper(pbar.update, f, "read")

                upload_response = requests.put(session_uri, headers=headers, data=upload_data, timeout=REQUEST_TIMEOUT)

                if upload_response.status_code in [200, 201]:
                    return response.token
                elif upload_response.status_code == 308:  # Resume Incomplete # noqa: PLR2004
                    uploaded_bytes = _check_uploaded_size(session_uri, file_size)
                else:
                    upload_failed_exception = (
                        f"Upload failed with status code {upload_response.status_code}: {upload_response.text}"
                    )
                    raise BackendError(upload_failed_exception)
            except (requests.ConnectionError, requests.Timeout) as e:
                logger.info(f"Network issue: {e}, retrying in {backoff_factor} seconds...")
                time.sleep(backoff_factor)
                backoff_factor = min(backoff_factor * 2, 60)
                retry_count += 1
                uploaded_bytes = _check_uploaded_size(session_uri, file_size)
                pbar.n = uploaded_bytes  # Update progress bar to reflect actual uploaded bytes

    return response.token


def upload_files_and_directories(
    folder: str,
    *,
    ignore_patterns: Sequence[str],
    item_type: ApiBlobType,
    quiet: bool = False,
) -> UploadDirectoryInfo:
    # Count the total number of files
    file_count = 0
    for _, _, files in filtered_walk(base_dir=folder, ignore_patterns=ignore_patterns):
        file_count += len(files)

    if file_count > MAX_FILES_TO_UPLOAD:
        if not quiet:
            logger.info(f"More than {MAX_FILES_TO_UPLOAD} files detected, creating a zip archive...")

        with TemporaryDirectory() as temp_dir:
            zip_path = os.path.join(temp_dir, TEMP_ARCHIVE_FILE)
            with zipfile.ZipFile(zip_path, "w") as zipf:
                for root, _, files in filtered_walk(base_dir=folder, ignore_patterns=ignore_patterns):
                    for file in files:
                        file_path = os.path.join(root, file)
                        zipf.write(file_path, os.path.relpath(file_path, folder))

            tokens = [
                token
                for token in [_upload_file(file_path=zip_path, item_type=item_type, quiet=quiet)]
                if token is not None
            ]
            return UploadDirectoryInfo(name="archive", files=tokens)

    root_dict = UploadDirectoryInfo(name="root")
    if os.path.isfile(folder):
        # Directly upload the file if the path is a file
        token = _upload_file(file_path=folder, item_type=item_type, quiet=quiet)
        if token:
            root_dict.files.append(token)
    else:
        for root, _, files in filtered_walk(base_dir=folder, ignore_patterns=ignore_patterns):
            # Path of the current folder relative to the base folder
            path = os.path.relpath(root, folder)

            # Navigate or create the dictionary path to the current folder
            current_dict = root_dict
            if path != ".":
                for part in path.split(os.sep):
                    # Find or create the subdirectory in the current dictionary
                    for subdir in current_dict.directories:
                        if subdir.name == part:
                            current_dict = subdir
                            break
                    else:
                        # If the directory is not found, create a new one
                        new_dir = UploadDirectoryInfo(name=part)
                        current_dict.directories.append(new_dir)
                        current_dict = new_dir

            # Add file tokens to the current directory in the dictionary
            for file in files:
                token = _upload_file(file_path=os.path.join(root, file), item_type=item_type, quiet=quiet)
                if token:
                    current_dict.files.append(token)

    return root_dict


def _upload_file(file_path: str, *, quiet: bool, item_type: ApiBlobType) -> str | None:
    """Helper function to upload a single file.

    Args:
        full_path: path to the file to upload
        quiet: suppress verbose output
        item_type: Type of the item that is being uploaded.

    Returns:
        A str token of uploaded file if successful, otherwise None.
    """

    if not quiet:
        logger.info("Starting upload for file " + file_path)

    if not os.path.isfile(file_path):
        logger.warn("Skip uploading %s because it is not a file.", file_path)
        return None

    content_length = os.path.getsize(file_path)
    token = _upload_blob(file_path, item_type)
    if not quiet:
        logger.info("Upload successful: " + file_path + " (" + File.get_size(content_length) + ")")
    return token


def normalize_patterns(*, default: list[str], additional: list[str] | str | None) -> list[str]:
    """Merges additional patterns with the default, and normalize the dir pattern with wildcard."""

    def add_wildcard_to_dir(pattern: str) -> str:
        return pattern + "*" if pattern.endswith("/") else pattern

    if additional is None:
        additional = []
    elif isinstance(additional, str):
        additional = [additional]

    return [add_wildcard_to_dir(pattern) for pattern in default + additional]



import unittest

from kagglesdk.datasets.types.dataset_enums import DatabundleVersionStatus
from kagglesdk.models.types.model_enums import ModelFramework

from kagglehub.enum import enum_to_str, to_enum


class TestEnumUtils(unittest.TestCase):
    def test_to_enum(self) -> None:
        # Enum values WITHOUT class prefix
        self.assertEqual(DatabundleVersionStatus.READY, to_enum(DatabundleVersionStatus, "ready"))
        self.assertEqual(
            DatabundleVersionStatus.INDIVIDUAL_BLOBS_COMPRESSED,
            to_enum(DatabundleVersionStatus, "individualBlobsCompressed"),
        )
        # Enum values WITH class prefix
        self.assertEqual(ModelFramework.MODEL_FRAMEWORK_JAX, to_enum(ModelFramework, "jax"))
        # self.assertEqual(ModelFramework.MODEL_FRAMEWORK_TENSOR_FLOW_2, to_enum(ModelFramework, "tensorFlow2"))
        self.assertEqual(ModelFramework.MODEL_FRAMEWORK_PY_TORCH, to_enum(ModelFramework, "pyTorch"))
        # Special case for PyTorch. "pyTorch" and "pytorch" are accepted
        self.assertEqual(ModelFramework.MODEL_FRAMEWORK_PY_TORCH, to_enum(ModelFramework, "pytorch"))

    def test_enum_to_str(self) -> None:
        # Enum WITHOUT class prefix
        self.assertEqual("ready", enum_to_str(DatabundleVersionStatus.READY))
        # Enum values WITH class prefix
        self.assertEqual("jax", enum_to_str(ModelFramework.MODEL_FRAMEWORK_JAX))
        self.assertEqual("tensorFlow2", enum_to_str(ModelFramework.MODEL_FRAMEWORK_TENSOR_FLOW_2))
        self.assertEqual("pyTorch", enum_to_str(ModelFramework.MODEL_FRAMEWORK_PY_TORCH))



import logging
import os
import time

from kagglesdk.kaggle_env import is_in_kaggle_notebook

from kagglehub.clients import (
    DEFAULT_CONNECT_TIMEOUT,
    KaggleJwtClient,
)
from kagglehub.config import is_kaggle_cache_disabled
from kagglehub.exceptions import BackendError
from kagglehub.handle import CompetitionHandle, DatasetHandle, ModelHandle, NotebookHandle
from kagglehub.logger import EXTRA_CONSOLE_BLOCK
from kagglehub.packages import PackageScope
from kagglehub.resolver import Resolver

KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME = "KAGGLE_CACHE_MOUNT_FOLDER"
ATTACH_DATASOURCE_REQUEST_NAME = "AttachDatasourceUsingJwtRequest"
# b/312965617: Using a longer timeout for this RPC.
ATTACH_DATASOURCE_READ_TIMEOUT = 30  # seconds

# TODO(b/385788821): this couples the client to backend implementation details
#   in AttachDatasourceUsingJwt. We should remove the need to prepend this in
#   kagglehub, and have AttachDatasourceUsingJwt return the actual mount path.
DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER = "/kaggle/input"


logger = logging.getLogger(__name__)


class CompetitionKaggleCacheResolver(Resolver[CompetitionHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        if is_kaggle_cache_disabled():
            return False
        if is_in_kaggle_notebook():
            return True
        return False

    def _resolve(
        self, h: CompetitionHandle, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        client = KaggleJwtClient()
        if force_download:
            logger.info(
                "Ignoring `force_download` argument when running inside the Kaggle notebook environment.",
                extra={**EXTRA_CONSOLE_BLOCK},
            )

        competition_ref = {
            "CompetitionSlug": h.competition,
        }
        result = client.post(
            ATTACH_DATASOURCE_REQUEST_NAME,
            {
                "competitionRef": competition_ref,
            },
            timeout=(DEFAULT_CONNECT_TIMEOUT, ATTACH_DATASOURCE_READ_TIMEOUT),
        )

        if "mountSlug" not in result:
            msg = "'result.mountSlug' field missing from response"
            raise BackendError(msg)

        base_mount_path = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME, DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER)
        cached_path = f"{base_mount_path}/{result['mountSlug']}"
        if not os.path.exists(cached_path):
            # Only print this if the competition is not already mounted.
            logger.info(f"Mounting files to {cached_path}...")
        elif path:
            logger.info(
                f"Attaching '{path}' from competition '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        else:
            logger.info(
                f"Attaching competition '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        while not os.path.exists(cached_path):
            time.sleep(5)
        if path:
            cached_filepath = f"{cached_path}/{path}"
            if not os.path.exists(cached_filepath):
                msg = (
                    f"'{path}' is not present in the competition files."
                    f"You can acces the other files othe attached competition at '{cached_path}'"
                )
                raise ValueError(msg)
            return cached_filepath, None
        return cached_path, None


class DatasetKaggleCacheResolver(Resolver[DatasetHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        if is_kaggle_cache_disabled():
            return False

        if is_in_kaggle_notebook():
            return True

        return False

    def _resolve(
        self, h: DatasetHandle, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        if force_download:
            logger.info(
                "Ignoring `force_download` argument when running inside the Kaggle notebook environment.",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        client = KaggleJwtClient()
        dataset_ref = {
            "OwnerSlug": h.owner,
            "DatasetSlug": h.dataset,
        }
        if h.is_versioned():
            dataset_ref["VersionNumber"] = str(h.version)
        else:
            # Check if there's a Package in scope which has stored a version number used when it was created.
            version_from_package_scope = PackageScope.get_version(h)
            if version_from_package_scope is not None:
                dataset_ref["VersionNumber"] = str(version_from_package_scope)

        result = client.post(
            ATTACH_DATASOURCE_REQUEST_NAME,
            {
                "datasetRef": dataset_ref,
            },
            timeout=(DEFAULT_CONNECT_TIMEOUT, ATTACH_DATASOURCE_READ_TIMEOUT),
        )
        if "mountSlug" not in result:
            msg = "'result.mountSlug' field missing from response"
            raise BackendError(msg)

        base_mount_path = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME, DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER)
        cached_path = f"{base_mount_path}/{result['mountSlug']}"
        version = result.get("versionNumber")  # None if missing

        if not os.path.exists(cached_path):
            # Only print this if the dataset is not already mounted.
            logger.info(f"Mounting files to {cached_path}...")
        elif path:
            logger.info(
                f"Attaching '{path}' from dataset '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        else:
            logger.info(
                f"Attaching dataset '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )

        while not os.path.exists(cached_path):
            time.sleep(5)

        if path:
            cached_filepath = f"{cached_path}/{path}"
            if not os.path.exists(cached_filepath):
                msg = (
                    f"'{path}' is not present in the dataset files."
                    f"You can acces the other files othe attached dataset at '{cached_path}'"
                )
                raise ValueError(msg)
            return cached_filepath, version
        return cached_path, version


class ModelKaggleCacheResolver(Resolver[ModelHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        if is_kaggle_cache_disabled():
            return False

        if is_in_kaggle_notebook():
            return True

        return False

    def _resolve(
        self, h: ModelHandle, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        if force_download:
            logger.info(
                "Ignoring `force_download` argument when running inside the Kaggle notebook environment.",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        client = KaggleJwtClient()
        model_ref = {
            "OwnerSlug": h.owner,
            "ModelSlug": h.model,
            "Framework": h.framework,
            "InstanceSlug": h.variation,
        }
        if h.is_versioned():
            model_ref["VersionNumber"] = str(h.version)
        else:
            # Check if there's a Package in scope which has stored a version number used when it was created.
            version_from_package_scope = PackageScope.get_version(h)
            if version_from_package_scope is not None:
                model_ref["VersionNumber"] = str(version_from_package_scope)

        result = client.post(
            ATTACH_DATASOURCE_REQUEST_NAME,
            {
                "modelRef": model_ref,
            },
            timeout=(DEFAULT_CONNECT_TIMEOUT, ATTACH_DATASOURCE_READ_TIMEOUT),
        )
        if "mountSlug" not in result:
            msg = "'result.mountSlug' field missing from response"
            raise BackendError(msg)

        base_mount_path = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME, DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER)
        cached_path = f"{base_mount_path}/{result['mountSlug']}"
        version = result.get("versionNumber")  # None if missing

        if not os.path.exists(cached_path):
            # Only print this if the model is not already mounted.
            logger.info(f"Mounting files to {cached_path}...")
        elif path:
            logger.info(
                f"Attaching '{path}' from model '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        else:
            logger.info(
                f"Attaching model '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )

        while not os.path.exists(cached_path):
            time.sleep(5)

        if path:
            cached_filepath = f"{cached_path}/{path}"
            if not os.path.exists(cached_filepath):
                msg = (
                    f"'{path}' is not present in the model files. "
                    f"You can access the other files of the attached model at '{cached_path}'"
                )
                raise ValueError(msg)
            return cached_filepath, version
        return cached_path, version


class NotebookOutputKaggleCacheResolver(Resolver[NotebookHandle]):
    def is_supported(self, *_, **__) -> bool:  # noqa: ANN002, ANN003
        if is_kaggle_cache_disabled():
            return False

        if is_in_kaggle_notebook():
            return True

        return False

    def _resolve(
        self, h: NotebookHandle, path: str | None = None, *, force_download: bool | None = False
    ) -> tuple[str, int | None]:
        if force_download:
            logger.info(
                "Ignoring `force_download` argument when running inside the Kaggle notebook environment.",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        client = KaggleJwtClient()
        kernel_ref = {
            "OwnerSlug": h.owner,
            "KernelSlug": h.notebook,
        }
        if h.is_versioned():
            kernel_ref["VersionNumber"] = str(h.version)
        else:
            # Check if there's a Package in scope which has stored a version number used when it was created.
            version_from_package_scope = PackageScope.get_version(h)
            if version_from_package_scope is not None:
                kernel_ref["VersionNumber"] = str(version_from_package_scope)

        result = client.post(
            ATTACH_DATASOURCE_REQUEST_NAME,
            {
                "kernelRef": kernel_ref,
            },
            timeout=(DEFAULT_CONNECT_TIMEOUT, ATTACH_DATASOURCE_READ_TIMEOUT),
        )

        if "mountSlug" not in result:
            msg = "'result.mountSlug' field missing from response"
            raise BackendError(msg)

        base_mount_path = os.getenv(KAGGLE_CACHE_MOUNT_FOLDER_ENV_VAR_NAME, DEFAULT_KAGGLE_CACHE_MOUNT_FOLDER)
        cached_path = f"{base_mount_path}/{result['mountSlug']}"
        version = result.get("versionNumber")  # None if missing

        if not os.path.exists(cached_path):
            # Only print this if the notebook output is not already mounted.
            logger.info(f"Mounting files to {cached_path}...")
        elif path:
            logger.info(
                f"Attaching '{path}' from Notebook '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )
        else:
            logger.info(
                f"Attaching model '{h}' to your Kaggle notebook...",
                extra={**EXTRA_CONSOLE_BLOCK},
            )

        while not os.path.exists(cached_path):
            time.sleep(5)

        if path:
            cached_filepath = f"{cached_path}/{path}"
            if not os.path.exists(cached_filepath):
                msg = (
                    f"'{path}' is not present in the notebook output files."
                    f"You can access the other files of the attached notebook at '{cached_path}'"
                )
                raise ValueError(msg)
            return cached_filepath, version
        return cached_path, version



import logging
from http import HTTPStatus

from kagglesdk.models.types.model_api_service import (
    ApiCreateModelInstanceRequest,
    ApiCreateModelInstanceRequestBody,
    ApiCreateModelInstanceVersionRequest,
    ApiCreateModelInstanceVersionRequestBody,
    ApiCreateModelRequest,
    ApiDeleteModelRequest,
    ApiGetModelRequest,
    CreateModelSigningTokenRequest,
)

from kagglehub.clients import BackendError, build_kaggle_client
from kagglehub.exceptions import KaggleApiHTTPError, handle_call, handle_mutate_call
from kagglehub.gcs_upload import UploadDirectoryInfo
from kagglehub.handle import ModelHandle

logger = logging.getLogger(__name__)


def _create_model(owner_slug: str, model_slug: str) -> None:
    with build_kaggle_client() as api_client:
        r = ApiCreateModelRequest()
        r.owner_slug = owner_slug
        r.slug = model_slug
        r.title = model_slug
        r.is_private = True
        handle_mutate_call(lambda: api_client.models.model_api_client.create_model(r))
        logger.info(f"Model '{model_slug}' Created.")


def _create_model_instance(
    model_handle: ModelHandle,
    upload_dir: UploadDirectoryInfo,
    license_name: str | None = None,
    *,
    sigstore: bool | None = False,
) -> None:
    upload_proto = upload_dir.to_proto()

    with build_kaggle_client() as api_client:
        r = ApiCreateModelInstanceRequest()
        r.owner_slug = model_handle.owner
        r.model_slug = model_handle.model
        r.body = ApiCreateModelInstanceRequestBody()
        r.body.instance_slug = model_handle.variation
        r.body.framework = model_handle.framework_enum()
        r.body.files = upload_proto.files
        r.body.directories = upload_proto.directories
        r.body.sigstore = sigstore
        if license_name is not None:
            r.body.license_name = license_name
        handle_mutate_call(lambda: api_client.models.model_api_client.create_model_instance(r))

        logger.info(
            f"Your model instance has been created.\nFiles are being processed...\nSee at: {model_handle.to_url()}"
        )


def _create_model_instance_version(
    model_handle: ModelHandle,
    upload_dir: UploadDirectoryInfo,
    version_notes: str = "",
    *,
    sigstore: bool | None = False,
) -> None:
    upload_proto = upload_dir.to_proto()

    with build_kaggle_client() as api_client:
        r = ApiCreateModelInstanceVersionRequest()
        r.owner_slug = model_handle.owner
        r.model_slug = model_handle.model
        r.framework = model_handle.framework_enum()
        r.instance_slug = model_handle.variation
        r.body = ApiCreateModelInstanceVersionRequestBody()
        r.body.version_notes = version_notes
        r.body.files = upload_proto.files
        r.body.directories = upload_proto.directories
        r.body.sigstore = sigstore
        handle_mutate_call(lambda: api_client.models.model_api_client.create_model_instance_version(r))
        logger.info(
            f"Your model instance version has been created.\n"
            f"Files are being processed...\nSee at: {model_handle.to_url()}"
        )


def create_model_instance_or_version(
    model_handle: ModelHandle,
    files: UploadDirectoryInfo,
    license_name: str | None,
    version_notes: str = "",
    *,
    sigstore: bool | None = False,
) -> None:
    try:
        _create_model_instance(model_handle, files, license_name, sigstore=sigstore)
    except BackendError as e:
        if e.error_code == HTTPStatus.CONFLICT:
            # Instance already exist, creating a new version instead.
            _create_model_instance_version(model_handle, files, version_notes, sigstore=sigstore)
        else:
            raise (e)


def create_model_if_missing(owner_slug: str, model_slug: str) -> None:
    try:
        with build_kaggle_client() as api_client:
            r = ApiGetModelRequest()
            r.owner_slug = owner_slug
            r.model_slug = model_slug
            handle_call(lambda: api_client.models.model_api_client.get_model(r))
    except KaggleApiHTTPError as e:
        if e.response is not None and (
            e.response.status_code == HTTPStatus.NOT_FOUND  # noqa: PLR1714
            or e.response.status_code == HTTPStatus.FORBIDDEN
        ):
            logger.info(
                f"Model '{model_slug}' does not exist or access is forbidden for user '{owner_slug}'. Creating or handling Model..."  # noqa: E501
            )
            _create_model(owner_slug, model_slug)
        else:
            raise (e)
    except Exception as e:
        raise e


def delete_model(owner_slug: str, model_slug: str) -> None:
    try:
        with build_kaggle_client() as api_client:
            r = ApiDeleteModelRequest()
            r.owner_slug = owner_slug
            r.model_slug = model_slug
            handle_mutate_call(lambda: api_client.models.model_api_client.delete_model(r))
    except KaggleApiHTTPError as e:
        if e.response is not None and e.response.status_code == HTTPStatus.NOT_FOUND:
            logger.info(f"Could not delete Model '{model_slug}' for user '{owner_slug}'...")
        else:
            raise (e)


def signing_token(owner_slug: str, model_slug: str) -> str | None:
    "Returns a JWT for signing if authorized for /{owner_slug}/{model_slug}"
    try:
        with build_kaggle_client() as api_client:
            r = CreateModelSigningTokenRequest()
            r.owner_slug = owner_slug
            r.model_slug = model_slug
            response = handle_call(lambda: api_client.models.model_api_client.create_model_signing_token(r))
            return response.id_token
    except KaggleApiHTTPError as e:
        if e.response is not None and e.response.status_code == HTTPStatus.NOT_FOUND:
            logger.info(
                f"Could not get Signing token for Model '{model_slug}' for user '{owner_slug}'. Skipping signing..."
            )
        return ""



from enum import Enum


class KaggleDatasetAdapter(Enum):
    HUGGING_FACE = "hugging_face"
    PANDAS = "pandas"
    POLARS = "polars"


# Enum to control how we load the file using polars. LazyFrames are the preferred
# way of working with polars: https://docs.pola.rs/api/python/stable/reference/lazyframe/index.html
# But users should be able to control whether they want a fully loaded DataFrame instead.
class PolarsFrameType(Enum):
    LAZY_FRAME = 1
    DATA_FRAME = 2



import logging
from http import HTTPStatus

from kagglesdk.datasets.types.dataset_api_service import (
    ApiCreateDatasetRequest,
    ApiCreateDatasetVersionRequest,
    ApiCreateDatasetVersionRequestBody,
    ApiDeleteDatasetRequest,
)

from kagglehub.clients import BackendError, build_kaggle_client
from kagglehub.exceptions import KaggleApiHTTPError, handle_mutate_call
from kagglehub.gcs_upload import UploadDirectoryInfo
from kagglehub.handle import DatasetHandle

logger = logging.getLogger(__name__)


def _create_dataset(dataset_handle: DatasetHandle, upload_dir: UploadDirectoryInfo) -> None:
    upload_proto = upload_dir.to_proto()

    with build_kaggle_client() as api_client:
        r = ApiCreateDatasetRequest()
        r.owner_slug = dataset_handle.owner
        r.slug = dataset_handle.dataset
        r.title = dataset_handle.dataset
        r.files = upload_proto.files
        r.directories = upload_proto.directories
        r.is_private = True
        handle_mutate_call(lambda: api_client.datasets.dataset_api_client.create_dataset(r))

        logger.info(f"Your dataset has been created.\nFiles are being processed...\nSee at: {dataset_handle.to_url()}")


def _create_dataset_version(
    dataset_handle: DatasetHandle, upload_dir: UploadDirectoryInfo, version_notes: str = ""
) -> None:
    upload_proto = upload_dir.to_proto()

    with build_kaggle_client() as api_client:
        r = ApiCreateDatasetVersionRequest()
        r.owner_slug = dataset_handle.owner
        r.dataset_slug = dataset_handle.dataset
        r.body = ApiCreateDatasetVersionRequestBody()
        r.body.version_notes = version_notes
        r.body.files = upload_proto.files
        r.body.directories = upload_proto.directories
        handle_mutate_call(lambda: api_client.datasets.dataset_api_client.create_dataset_version(r))
        logger.info(
            f"Your dataset version has been created.\nFiles are being processed...\nSee at: {dataset_handle.to_url()}"
        )


def create_dataset_or_version(
    dataset_handle: DatasetHandle, files: UploadDirectoryInfo, version_notes: str = ""
) -> None:
    try:
        _create_dataset(dataset_handle, files)
    except BackendError as e:
        if e.error_code in (None, HTTPStatus.CONFLICT):
            # Dataset already exists, creating a new version instead.
            _create_dataset_version(dataset_handle, files, version_notes)
        else:
            raise (e)


def delete_dataset(owner_slug: str, dataset_slug: str) -> None:
    try:
        with build_kaggle_client() as api_client:
            r = ApiDeleteDatasetRequest()
            r.owner_slug = owner_slug
            r.dataset_slug = dataset_slug
            handle_mutate_call(lambda: api_client.datasets.dataset_api_client.delete_dataset(r))
    except KaggleApiHTTPError as e:
        if e.response is not None and e.response.status_code == HTTPStatus.NOT_FOUND:
            logger.info(f"Could not delete Dataset '{dataset_slug}' for user '{owner_slug}'...")
        else:
            raise (e)



import threading
from dataclasses import dataclass

from flask import Flask, jsonify, request
from flask.typing import ResponseReturnValue
from kagglesdk.datasets.types.dataset_api_service import ApiDataset, ApiDownloadDatasetRequest, ApiGetDatasetRequest

from tests.utils import AUTO_COMPRESSED_FILE_NAME, add_mock_gcs_route, get_gcs_redirect_response

app = Flask(__name__)
add_mock_gcs_route(app)

TARGZ_ARCHIVE_HANDLE = "testuser/zip-dataset/versions/1"


@dataclass
class SharedData:
    last_download_user_agent = ""


shared_data: SharedData = SharedData()
lock = threading.Lock()


@app.route("/", methods=["HEAD"])
def head() -> ResponseReturnValue:
    return "", 200


@app.route("/api/v1/datasets.DatasetApiService/GetDataset", methods=["POST"])
def dataset_get() -> ResponseReturnValue:
    r = ApiGetDatasetRequest.from_dict(request.get_json())
    dataset = ApiDataset()
    dataset.owner_name = r.owner_slug
    dataset.title = r.dataset_slug
    dataset.current_version_number = 2
    return dataset.to_json(), 200


# For Datasets, downloads of the archive and individual files happen at the same route, controlled
# by a file_name query param
@app.route("/api/v1/datasets.DatasetApiService/DownloadDataset", methods=["POST"])
def dataset_download() -> ResponseReturnValue:
    lock.acquire()
    shared_data.last_download_user_agent = request.headers.get("User-Agent", "")
    lock.release()

    r = ApiDownloadDatasetRequest.from_dict(request.get_json())

    handle = f"{r.owner_slug}/{r.dataset_slug}"

    # First, determine if we're fetching a file or the whole dataset
    if r.file_name:
        # This mimics behavior for our file downloads, where users request a file, but
        # receive a zipped version of the file from GCS.
        test_file_name = f"{AUTO_COMPRESSED_FILE_NAME}.zip" if r.file_name == AUTO_COMPRESSED_FILE_NAME else r.file_name
    # Check a special case to handle tar.gz
    elif handle in TARGZ_ARCHIVE_HANDLE:
        test_file_name = "archive.tar.gz"
    else:
        test_file_name = "foo.txt.zip"

    return get_gcs_redirect_response(test_file_name)


@app.errorhandler(404)
def error(e: Exception):  # noqa: ANN201
    data = {"message": "Some response data", "error": str(e)}
    return jsonify(data), 404
